"""These tests rely on the pytest tmp_path factory fixture"""

import os
from pathlib import Path

import pytest
from metricflow.dataflow.sql_column import SqlColumn

from ruamel.yaml import YAML

from metricflow.inference.models import InferenceResult, InferenceSignalType
from metricflow.inference.renderer.config_file import ConfigFileRenderer

yaml = YAML()


def test_no_overwrite_with_existing_dir(tmpdir: Path) -> None:  # noqa: D
    with pytest.raises(ValueError):
        ConfigFileRenderer(tmpdir, False)


def test_dir_path_is_file(tmpdir: Path) -> None:  # noqa: D
    file_path = os.path.join(tmpdir, "file.txt")
    with open(file_path, "w") as f:
        f.write("file contents!")

    with pytest.raises(ValueError):
        ConfigFileRenderer(file_path, False)


def test_render_configs(tmpdir: Path) -> None:  # noqa: D
    inference_results = [
        InferenceResult(
            column=SqlColumn.from_string("db.schema.test_table.id"),
            type_node=InferenceSignalType.ID.PRIMARY,
            reasons=[],
            problems=[],
        ),
        InferenceResult(
            column=SqlColumn.from_string("db.schema.test_table.time_dim"),
            type_node=InferenceSignalType.DIMENSION.TIME,
            reasons=[],
            problems=[],
        ),
        InferenceResult(
            column=SqlColumn.from_string("db.schema.test_table.primary_time_dim"),
            type_node=InferenceSignalType.DIMENSION.PRIMARY_TIME,
            reasons=[],
            problems=[],
        ),
    ]

    renderer = ConfigFileRenderer(tmpdir, True)

    renderer.render(inference_results)

    table_file_path = os.path.join(tmpdir, "db.schema.test_table.yaml")
    assert os.path.isfile(table_file_path)

    with open(table_file_path, "r") as f:
        file_contents = yaml.load(f)

    assert file_contents == {
        "data_source": {
            "name": "test_table",
            "sql_table": "db.schema.test_table",
            "identifiers": [{"type": "primary", "name": "id"}],
            "dimensions": [
                {"type": "time", "name": "time_dim", "type_params": {"time_granularity": "day"}},
                {
                    "type": "time",
                    "name": "primary_time_dim",
                    "type_params": {"is_primary": True, "time_granularity": "day"},
                },
            ],
            "measures": [],
        }
    }



from __future__ import annotations
from abc import abstractmethod

from typing import Callable, List, TypeVar

from metricflow.inference.context.data_warehouse import (
    ColumnProperties,
    DataWarehouseInferenceContext,
)
from metricflow.inference.rule.base import InferenceRule
from metricflow.inference.models import (
    InferenceSignal,
    InferenceSignalConfidence,
    InferenceSignalNode,
)

T = TypeVar("T", bound="ColumnMatcherRule")

ColumnMatcher = Callable[[T, ColumnProperties], bool]


class ColumnMatcherRule(InferenceRule):
    """Inference rule that checks for matches across all columns.

    This is a useful shortcut for making rules that match columns one by one with preset confidence
    values, types and match reasons.

    If you need a more specific rule with varying confidence, column cross-checking and that outputs
    multiple types, inherit from `InferenceRule` directly.

    type_node: the `InferenceSignalNode` to produce whenever the pattern is matched
    confidence: the `InferenceSignalConfidence` to produce whenever the pattern is matched
    only_applies_to_parent_signal: whether the produced signal should be only taken into
        consideration by the solver if the parent is present in the tree.
    match_reason: a human-readable string of the reason why this was matched
    """

    type_node: InferenceSignalNode
    confidence: InferenceSignalConfidence
    only_applies_to_parent_signal: bool
    match_reason: str

    @abstractmethod
    def match_column(self, props: ColumnProperties) -> bool:
        """A function to determine whether `ColumnProperties` matches. If it does, produce the signal"""
        raise NotImplementedError

    def process(self, warehouse: DataWarehouseInferenceContext) -> List[InferenceSignal]:  # type: ignore
        """Try to match all columns' properties with the matching function.

        If they do match, produce a signal with the configured type and confidence.
        """
        matching_columns = [column for column, props in warehouse.columns.items() if self.match_column(props)]
        signals = [
            InferenceSignal(
                column=column,
                type_node=self.type_node,
                reason=self.match_reason,
                confidence=self.confidence,
                only_applies_to_parent=self.only_applies_to_parent_signal,
            )
            for column in matching_columns
        ]
        return signals


class LowCardinalityRatioRule(ColumnMatcherRule):
    """Inference rule that checks for string columns with low cardinality to count ratio.

    The ratio is calculated as `distinct_count/(count - null_count)`.
    """

    def __init__(self, cardinality_count_ratio_threshold: float) -> None:
        """Initialize the rule.

        cardinality_count_ratio_threshold: rations below this threshold will match.
        """
        assert cardinality_count_ratio_threshold >= 0 and cardinality_count_ratio_threshold <= 1
        self.threshold = cardinality_count_ratio_threshold
        super().__init__()

    match_reason = "Column has low cardinality"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        denom = props.row_count - props.null_count

        # undefined ratio
        if denom == 0:
            return False

        ratio = props.distinct_row_count / denom
        return ratio < self.threshold



import textwrap

from metricflow.dataflow.sql_table import SqlTable

from metricflow.model.objects.common import YamlConfigFile
from metricflow.model.objects.materialization import (
    MaterializationFormat,
    MaterializationLocation,
    MaterializationTableauParams,
)
from metricflow.model.parsing.dir_to_model import parse_yaml_files_to_model
from metricflow.model.validations.validator_helpers import ModelValidationException


def test_simple_materialization_parsing() -> None:
    """Test for parsing a simple maeterialization specification with no optional parameters"""
    yaml_contents = textwrap.dedent(
        """\
        materialization:
          name: simple_materialization_test
          metrics:
            - some_metric
          dimensions:
            - some_dimension
            - time_dimension
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.materializations) == 1
    materialization = build_result.model.materializations[0]
    assert materialization.name == "simple_materialization_test"
    assert materialization.metrics == ["some_metric"]
    assert materialization.dimensions == ["some_dimension", "time_dimension"]


def test_materialization_metadata_parsing() -> None:
    """Test for asserting that internal metadata is parsed into the Materialization object"""
    yaml_contents = textwrap.dedent(
        """\
        materialization:
          name: simple_materialization_test
          metrics:
            - some_metric
          dimensions:
            - some_dimension
            - time_dimension
        """
    )
    file = YamlConfigFile(filepath="test_dir/inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.materializations) == 1
    materialization = build_result.model.materializations[0]
    assert materialization.metadata is not None
    assert materialization.metadata.repo_file_path == "test_dir/inline_for_test"
    assert materialization.metadata.file_slice.filename == "inline_for_test"
    expected_metadata_content = textwrap.dedent(
        """\
        name: simple_materialization_test
        metrics:
        - some_metric
        dimensions:
        - some_dimension
        - time_dimension
        """
    )
    assert materialization.metadata.file_slice.content == expected_metadata_content


def test_materialization_with_simple_destinations_parsing() -> None:
    """Test for parsing a materialization specification with destinations set"""
    yaml_contents = textwrap.dedent(
        """\
        materialization:
          name: materialization_simple_destinations_test
          metrics:
            - some_metric
          dimensions:
            - some_dimension
          destinations:
            - location: dw
              format: wide
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.materializations) == 1
    materialization = build_result.model.materializations[0]
    assert materialization.destinations is not None
    assert len(materialization.destinations) == 1
    destination = materialization.destinations[0]
    assert destination.location is MaterializationLocation.DW
    assert destination.format is MaterializationFormat.WIDE


def test_materialization_with_rollup_destination_parsing() -> None:
    """Test for parsing a materialization specification with a rollup set on a destination"""
    yaml_contents = textwrap.dedent(
        """\
        materialization:
          name: materialization_simple_destinations_test
          metrics:
            - some_metric
          dimensions:
            - some_dimension
          destinations:
            - location: dw
              format: wide
              rollups:
               - []
               - ["some_dimension", "*"]
               - ["some_dimension"]
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.materializations) == 1
    materialization = build_result.model.materializations[0]
    assert materialization.destinations is not None
    assert len(materialization.destinations) == 1
    destination = materialization.destinations[0]
    assert destination.rollups is not None
    assert destination.rollups == [[], ["some_dimension", "*"], ["some_dimension"]]


def test_materialization_with_indented_rollup_destination_parsing() -> None:
    """Test for parsing a materialization specification with a rollup set on a destination"""
    yaml_contents = textwrap.dedent(
        """\
        materialization:
          name: materialization_simple_destinations_test
          metrics:
            - some_metric
          dimensions:
            - some_dimension
          destinations:
            - location: dw
              format: wide
              rollups:
               -
                 - "some_dimension"
                 - "*"
               -
                 - "some_dimension"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.materializations) == 1
    materialization = build_result.model.materializations[0]
    assert materialization.destinations is not None
    assert len(materialization.destinations) == 1
    destination = materialization.destinations[0]
    assert destination.rollups is not None
    assert destination.rollups == [["some_dimension", "*"], ["some_dimension"]]


def test_materialization_with_tableau_parameters_parsing() -> None:
    """Test for parsing a materialization specification with Tableau parameters"""
    yaml_contents = textwrap.dedent(
        """\
        materialization:
          name: materialization_simple_destinations_test
          metrics:
            - some_metric
          dimensions:
            - some_dimension
          destinations:
            - location: dw
              format: wide
              tableau_params:
                projects:
                 - some_project
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.materializations) == 1
    materialization = build_result.model.materializations[0]
    assert materialization.destinations is not None
    assert len(materialization.destinations) == 1
    destination = materialization.destinations[0]
    assert destination.tableau_params == MaterializationTableauParams(projects=["some_project"])


def test_materialization_with_destination_table_parsing() -> None:
    """Test for parsing a materialization specification with a destination table set"""
    yaml_contents = textwrap.dedent(
        """\
        materialization:
          name: materialization_simple_destinations_test
          metrics:
            - some_metric
          dimensions:
            - some_dimension
          destinations:
            - location: dw
              format: wide
          destination_table: "bq_project.output_schema.materialization_table"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.materializations) == 1
    materialization = build_result.model.materializations[0]
    assert materialization.destinations is not None and len(materialization.destinations) == 1
    assert materialization.destination_table is not None
    assert materialization.destination_table == SqlTable(
        db_name="bq_project", schema_name="output_schema", table_name="materialization_table"
    )


def test_materialization_with_invalid_destination_table_parsing_error() -> None:
    """Test for parsing error thrown when the sql table name cannot be properly parsed"""
    yaml_contents = textwrap.dedent(
        """\
        materialization:
          name: materialization_simple_destinations_test
          metrics:
            - some_metric
          dimensions:
            - some_dimension
          destinations:
            - location: dw
              format: wide
          destination_table: "this is an invalid table name"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])
    assert build_result.issues.has_blocking_issues
    assert "Invalid input for a SQL table" in str(ModelValidationException(build_result.issues.all_issues))



import logging

import pytest

from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.errors.errors import UnableToSatisfyQueryError
from metricflow.query.query_parser import MetricFlowQueryParser
from metricflow.specs import (
    MetricSpec,
    DimensionSpec,
    TimeDimensionSpec,
    IdentifierSpec,
    OrderBySpec,
    OutputColumnNameOverride,
)
from metricflow.test.test_utils import as_datetime
from metricflow.test.time.metric_time_dimension import MTD
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


def test_query_parser(query_parser: MetricFlowQueryParser) -> None:  # noqa: D
    query_spec = query_parser.parse_and_validate_query(
        metric_names=["bookings"], group_by_names=["is_instant", "listing", MTD], order=[MTD, "-bookings"]
    )

    assert query_spec.metric_specs == (MetricSpec(element_name="bookings"),)
    assert query_spec.dimension_specs == (DimensionSpec(element_name="is_instant", identifier_links=()),)
    assert query_spec.time_dimension_specs == (
        TimeDimensionSpec(element_name=MTD, identifier_links=(), time_granularity=TimeGranularity.DAY),
    )
    assert query_spec.identifier_specs == (IdentifierSpec(element_name="listing", identifier_links=()),)
    assert query_spec.order_by_specs == (
        OrderBySpec(
            item=TimeDimensionSpec(element_name=MTD, identifier_links=(), time_granularity=TimeGranularity.DAY),
            descending=False,
        ),
        OrderBySpec(
            item=MetricSpec(element_name="bookings"),
            descending=True,
        ),
    )


def test_order_by_granularity_conversion(query_parser: MetricFlowQueryParser) -> None:
    """Test that the granularity of the primary time dimension in the order by is returned appropriately.

    In the case where the primary time dimension is specified in the order by without a granularity suffix, the order
    by spec returned by the parser should have a granularity appropriate for the queried metrics.
    """

    # "bookings" has a granularity of DAY, "revenue" has a granularity of MONTH
    query_spec = query_parser.parse_and_validate_query(
        metric_names=["bookings", "revenue"], group_by_names=[MTD], order=[f"-{MTD}"]
    )

    # The lowest common granularity is MONTH, so we expect the PTD in the order by to have that granularity.
    assert (
        OrderBySpec(
            item=TimeDimensionSpec(element_name=MTD, identifier_links=(), time_granularity=TimeGranularity.DAY),
            descending=True,
        ),
    ) == query_spec.order_by_specs


def test_order_by_granularity_no_conversion(query_parser: MetricFlowQueryParser) -> None:  # noqa: D
    query_spec = query_parser.parse_and_validate_query(metric_names=["bookings"], group_by_names=[MTD], order=[MTD])

    # The only granularity is DAY, so we expect the PTD in the order by to have that granularity.
    assert (
        OrderBySpec(
            item=TimeDimensionSpec(element_name=MTD, identifier_links=(), time_granularity=TimeGranularity.DAY),
            descending=False,
        ),
    ) == query_spec.order_by_specs


def test_time_range_constraint_conversion(query_parser: MetricFlowQueryParser) -> None:
    """Test that the returned time constraint in the query spec is adjusted to match the granularity of the query."""

    # "bookings" has a granularity of DAY, "revenue" has a granularity of MONTH
    query_spec = query_parser.parse_and_validate_query(
        metric_names=["bookings", "revenue"],
        group_by_names=[MTD],
        time_constraint_start=as_datetime("2020-01-15"),
        time_constraint_end=as_datetime("2020-02-15"),
    )

    assert (
        TimeRangeConstraint(start_time=as_datetime("2020-01-15"), end_time=as_datetime("2020-02-15"))
    ) == query_spec.time_range_constraint


def test_column_override(query_parser: MetricFlowQueryParser) -> None:
    """Tests that the output column override is set.

    Should be set in cases where the metrics have a non-day granularity, but ds is specified.
    """

    # "revenue" has a granularity of MONTH
    query_spec = query_parser.parse_and_validate_query(
        metric_names=["revenue"],
        group_by_names=[MTD],
        time_constraint_start=as_datetime("2020-01-15"),
        time_constraint_end=as_datetime("2020-02-15"),
    )

    assert (
        OutputColumnNameOverride(
            time_dimension_spec=TimeDimensionSpec(
                element_name=MTD,
                identifier_links=(),
                time_granularity=TimeGranularity.DAY,
            ),
            output_column_name=MTD,
        ),
    ) == query_spec.output_column_name_overrides


def test_parse_and_validate_where_constraint_dims(query_parser: MetricFlowQueryParser) -> None:
    """Test that the returned time constraint in the query spec is adjusted to match the granularity of the query."""
    # check constraint on invalid_dim raises UnableToSatisfyQueryError
    with pytest.raises(UnableToSatisfyQueryError):
        query_parser.parse_and_validate_query(
            metric_names=["bookings"],
            group_by_names=[MTD],
            time_constraint_start=as_datetime("2020-01-15"),
            time_constraint_end=as_datetime("2020-02-15"),
            where_constraint_str="WHERE invalid_dim = '1'",
        )

    query_spec = query_parser.parse_and_validate_query(
        metric_names=["bookings"],
        group_by_names=[MTD],
        time_constraint_start=as_datetime("2020-01-15"),
        time_constraint_end=as_datetime("2020-02-15"),
        where_constraint_str="WHERE is_instant = '1'",
    )
    assert DimensionSpec(element_name="is_instant", identifier_links=()) not in query_spec.dimension_specs


def test_parse_and_validate_metric_constraint_dims(query_parser: MetricFlowQueryParser) -> None:
    """Test that the returned time constraint in the query spec is adjusted to match the granularity of the query."""

    # check constraint on invalid_dim raises UnableToSatisfyQueryError
    with pytest.raises(UnableToSatisfyQueryError):
        query_parser.parse_and_validate_query(
            metric_names=["metric_with_invalid_constraint"],
            group_by_names=[MTD],
            time_constraint_start=as_datetime("2020-01-15"),
            time_constraint_end=as_datetime("2020-02-15"),
        )



from __future__ import annotations

import logging
import re
from typing import List, Optional, Dict, Any

from moz_sql_parser import parse as moz_parse

from metricflow.errors.errors import ConstraintParseException
from metricflow.model.objects.base import HashableBaseModel, PydanticCustomInputParser, PydanticParseableValueType
from metricflow.sql.sql_bind_parameters import SqlBindParameters

logger = logging.getLogger(__name__)

LITERAL_STR = "literal"
INTERVAL_LITERAL = "interval"


class WhereClauseConstraint(PydanticCustomInputParser, HashableBaseModel):
    """Contains a string that is a where clause"""

    where: str
    linkable_names: List[str]
    sql_params: SqlBindParameters

    def __init__(  # noqa: D
        self,
        where: str = "",
        linkable_names: Optional[List[str]] = None,
        sql_params: Optional[SqlBindParameters] = None,
        # sql params: user-originated sql params that need to be escaped in a dialect-specific way keys are the
        # name of the template value in the `where` string, values are the string to be escaped and
        # inserted into the where string (ie where = "%(1)s", sql_values = {"1": "cote d'ivoire"})
    ) -> None:
        where = where.strip("\n") if where else ""
        linkable_names = linkable_names or []
        if sql_params is None:
            sql_params = SqlBindParameters()
        super().__init__(
            where=where,
            linkable_names=linkable_names,
            sql_params=sql_params,
        )

    @classmethod
    def _from_yaml_value(cls, input: PydanticParseableValueType) -> WhereClauseConstraint:
        """Parses a WhereClauseConstraint from a constraint string found in a user-provided model specification

        User-provided constraint strings are SQL snippets conforming to the expectations of SQL WHERE clauses,
        and as such we parse them using our standard parse method below.
        """
        if isinstance(input, str):
            return WhereClauseConstraint.parse(input)
        else:
            raise ValueError(f"Expected input to be of type string, but got type {type(input)} with value: {input}")

    @staticmethod
    def parse(s: str) -> WhereClauseConstraint:
        """Parse a string into a WhereClauseConstraint

        We are assuming here that if we needed to parse a string, we wouldn't have bind parameters.
        Because if we had bind-parameters, the string would have existing structure, and we wouldn't need to parse it.
        """
        s = strip_where(s)

        where_str = f"WHERE {s}"
        # to piggyback on moz sql parser we need a SELECT statement
        # moz breaks the sql statement into clauses:
        # where_str = "WHERE is_instant" yields -> {'select': {'value': '_'}, 'from': '_', 'where': 'is_instant'}
        # where_str = "WHERE is_instant AND country = 'vanuatu' AND is_lux or ds < '2020-01-02'" yields ->
        # {'select': {'value': '_'}, 'from': '_', 'where': {'or': [{'and': ['is_instant', {'eq': ['country', {'literal': 'vanuatu'}]}, 'is_lux']}, {'lt': ['ds', {'literal': '2020-01-02'}]}]}}
        parsed = moz_parse(f"select _ from _ {where_str}")
        if "where" not in parsed:
            raise ConstraintParseException(parsed)

        where = parsed["where"]
        if isinstance(where, dict):
            if not len(where.keys()) == 1:
                raise ConstraintParseException(f"expected parsed constraint to contain exactly one key; got {where}")
            return WhereClauseConstraint(
                where=s,
                linkable_names=constraint_dimension_names_from_dict(where),
                sql_params=SqlBindParameters(),
            )
        elif isinstance(where, str):
            return WhereClauseConstraint(
                where=s,
                linkable_names=[where.strip()],
                sql_params=SqlBindParameters(),
            )
        else:
            raise TypeError(f"where-clause is neither a dict nor a string. Unexpectedly it is a {type(where)}")

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}" f"(where={self.where}, linkable_names={self.linkable_names})"


def strip_where(s: str) -> str:
    """Removes WHERE from the beginning of the string, if present (regardless of case)"""
    # '^' tells the regex to only check the beginning of the string
    return re.sub("^where ", "", s, flags=re.IGNORECASE)


def constraint_dimension_names_from_dict(where: Dict[str, Any]) -> List[str]:  # type: ignore[misc] # noqa: D
    dims = []
    for key, clause in where.items():
        if key == LITERAL_STR or key == INTERVAL_LITERAL:
            continue
        dims += _get_dimensions_from_clause(clause)

    return dims


def constraint_values_from_dict(where: Dict[str, Any]) -> List[str]:  # type: ignore[misc] # noqa: d
    values = []
    for key, clause in where.items():
        if key == LITERAL_STR:
            values.append(clause)
        elif isinstance(clause, dict):
            values += constraint_values_from_dict(clause)
        elif isinstance(clause, list):
            for item in clause:
                if isinstance(item, dict):
                    values += constraint_values_from_dict(item)

    return values


def _constraint_dimensions_from_list(list_clause: List[Any]) -> List[str]:  # type: ignore[misc] # noqa: D
    dims = []
    for clause in list_clause:
        dims += _get_dimensions_from_clause(clause)

    return dims


def _get_dimensions_from_clause(clause: Any) -> List[str]:  # type: ignore[misc] # noqa: D
    if clause is not None:
        if isinstance(clause, dict):
            return constraint_dimension_names_from_dict(clause)
        elif isinstance(clause, list):
            return _constraint_dimensions_from_list(clause)
        elif isinstance(clause, str):
            return [clause.strip()]

    return []



# Need to ignore flake8 as there are blank lines in the test outputs, which throws "W293 blank line contains whitespace"
# flake8: noqa

import logging
import textwrap

import pytest

from metricflow.errors.errors import UnableToSatisfyQueryError
from metricflow.query.query_parser import MetricFlowQueryParser

logger = logging.getLogger(__name__)


def test_nonexistent_metric(query_parser: MetricFlowQueryParser) -> None:  # noqa: D
    with pytest.raises(UnableToSatisfyQueryError) as exception_info:
        query_parser.parse_and_validate_query(metric_names=["booking"], group_by_names=["is_instant"])

    assert (
        textwrap.dedent(
            """\
            Unable To Satisfy Query Error: Unknown metric: 'booking'
        
            Suggestions for 'booking':
                ['bookings',
                 'booking_fees',
                 'booking_value',
                 'instant_bookings',
                 'booking_payments',
                 'max_booking_value']
            """
        ).rstrip()
        == str(exception_info.value)
    )


def test_non_existent_group_by(query_parser: MetricFlowQueryParser) -> None:  # noqa: D
    with pytest.raises(UnableToSatisfyQueryError) as exception_info:
        query_parser.parse_and_validate_query(metric_names=["bookings"], group_by_names=["is_instan"])

    assert (
        textwrap.dedent(
            """\
            Unable To Satisfy Query Error: Unknown element name 'is_instan' in dimension name 'is_instan'
            
            Suggestions for 'is_instan':
                ['is_instant']
            """
        ).rstrip()
        == str(exception_info.value)
    )


def test_invalid_group_by(query_parser: MetricFlowQueryParser) -> None:  # noqa: D
    with pytest.raises(UnableToSatisfyQueryError) as exception_info:
        query_parser.parse_and_validate_query(metric_names=["bookings"], group_by_names=["capacity_latest"])

    assert (
        textwrap.dedent(
            """\
            Unable To Satisfy Query Error: Dimensions ['capacity_latest'] cannot be resolved for metrics \
['bookings']. The invalid dimension may not exist, require an ambiguous join (e.g. a join path that can be satisfied \
in multiple ways), or require a fanout join.
            
            Suggestions for invalid dimension 'capacity_latest':
                ['listing__capacity_latest', 'listing__country_latest']
            """
        ).rstrip()
        == str(exception_info.value)
    )



from _pytest.fixtures import FixtureRequest

from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer
from metricflow.sql.sql_plan import SqlQueryPlan, SqlQueryPlanNode, SqlSelectStatementNode
from metricflow.sql.sql_plan_to_text import sql_query_plan_as_text
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.plan_utils import (
    assert_plan_snapshot_text_equal,
    make_schema_replacement_function,
)


def assert_default_rendered_sql_equal(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    plan_id: str,
    sql_plan_node: SqlQueryPlanNode,
) -> None:
    """Helper function to render a select statement and compare with the one saved as a file."""
    sql_query_plan = SqlQueryPlan(plan_id=plan_id, render_node=sql_plan_node)

    rendered_sql = DefaultSqlQueryPlanRenderer().render_sql_query_plan(sql_query_plan).sql

    assert_plan_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan=sql_query_plan,
        plan_snapshot_text=rendered_sql,
        plan_snapshot_file_extension=".sql",
    )


def assert_rendered_sql_equal(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    plan_id: str,
    select_node: SqlSelectStatementNode,
    sql_client: SqlClient,
) -> None:
    """Helper function to render a select statement and compare with the one saved as a file."""
    sql_query_plan = SqlQueryPlan(plan_id=plan_id, render_node=select_node)

    assert_rendered_sql_from_plan_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_query_plan=sql_query_plan,
        sql_client=sql_client,
    )


def assert_rendered_sql_from_plan_equal(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_query_plan: SqlQueryPlan,
    sql_client: SqlClient,
) -> None:
    """Similar to assert_rendered_sql_equal, but takes in a SQL query plan"""
    rendered_sql = sql_client.sql_engine_attributes.sql_query_plan_renderer.render_sql_query_plan(sql_query_plan).sql

    assert_plan_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan=sql_query_plan,
        plan_snapshot_text=rendered_sql,
        plan_snapshot_file_extension=".sql",
        incomparable_strings_replacement_function=make_schema_replacement_function(
            system_schema=mf_test_session_state.mf_system_schema, source_schema=mf_test_session_state.mf_source_schema
        ),
        additional_sub_directories_for_snapshots=(sql_client.__class__.__name__,) if sql_client else (),
    )


def assert_sql_plan_text_equal(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_query_plan: SqlQueryPlan,
) -> None:
    assert_plan_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan=sql_query_plan,
        plan_snapshot_text=sql_query_plan_as_text(sql_query_plan),
        incomparable_strings_replacement_function=make_schema_replacement_function(
            system_schema=mf_test_session_state.mf_system_schema, source_schema=mf_test_session_state.mf_source_schema
        ),
    )



from abc import ABC, abstractmethod
from typing import List

from metricflow.dataflow.sql_column import SqlColumn
from metricflow.inference.models import InferenceResult, InferenceSignal


class InferenceSolver(ABC):
    """Base class for inference solvers.

    Inference solvers implement algorithms for making a final decision over what a column is or isn't, based on
    signals produced by inference rules for that column.
    """

    @abstractmethod
    def solve_column(self, column: SqlColumn, signals: List[InferenceSignal]) -> InferenceResult:
        """Make a decision about a column based on its input signals."""
        raise NotImplementedError



from metricflow.protocols.sql_client import SqlClient
from metricflow.dataflow.sql_table import SqlTable
from metricflow.execution.execution_plan import (
    ExecutionPlan,
    SelectSqlQueryToDataFrameTask,
    SelectSqlQueryToTableTask,
)
from metricflow.execution.executor import SequentialPlanExecutor
from metricflow.object_utils import random_id
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_clients.sql_utils import make_df
from metricflow.test.compare_df import assert_dataframes_equal
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState


def test_read_sql_task(sql_client: SqlClient) -> None:  # noqa: D
    task = SelectSqlQueryToDataFrameTask(sql_client, "SELECT 1 AS foo", SqlBindParameters())
    execution_plan = ExecutionPlan("plan0", leaf_tasks=[task])

    results = SequentialPlanExecutor().execute_plan(execution_plan)
    task_result = results.get_result(task.task_id)

    assert not results.contains_task_errors
    assert task_result.df is not None

    assert_dataframes_equal(
        actual=task_result.df,
        expected=make_df(
            sql_client=sql_client,
            columns=["foo"],
            data=[(1,)],
        ),
    )


def test_write_table_task(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> None:  # noqa: D
    output_table = SqlTable(schema_name=mf_test_session_state.mf_system_schema, table_name=f"test_table_{random_id()}")
    task = SelectSqlQueryToTableTask(
        sql_client=sql_client,
        sql_query="SELECT 1 AS foo",
        execution_parameters=SqlBindParameters(),
        output_table=output_table,
    )
    execution_plan = ExecutionPlan("plan0", leaf_tasks=[task])

    results = SequentialPlanExecutor().execute_plan(execution_plan)

    assert not results.contains_task_errors
    assert sql_client.table_exists(output_table)

    assert_dataframes_equal(
        actual=sql_client.query(f"SELECT * FROM {output_table.sql}"),
        expected=make_df(
            sql_client=sql_client,
            columns=["foo"],
            data=[(1,)],
        ),
    )
    sql_client.drop_table(output_table)



from abc import ABC, abstractmethod
import contextlib
from dataclasses import InitVar, dataclass, field
from datetime import date, datetime
from enum import Enum
from typing import Callable, ContextManager, Dict, Iterator, List, Optional, TypeVar, Generic

from metricflow.dataflow.sql_column import SqlColumn
from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SqlClient
from metricflow.inference.context.base import InferenceContext, InferenceContextProvider

T = TypeVar("T", str, int, float, date, datetime)


class InferenceColumnType(str, Enum):
    """Represents a column type that can be used for inference.

    This does not provide a 1 to 1 mapping between SQL types and enum values. For example,
    all possible floating point types (FLOAT, DOUBLE etc) are mapped to the same FLOAT
    value. Same for datetimes and others.
    """

    STRING = "string"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    FLOAT = "float"
    DATETIME = "datetime"
    UNKNOWN = "unknown"


@dataclass(frozen=True)
class ColumnProperties(Generic[T]):
    """Holds properties about a column that were extracted from the data warehouse."""

    column: SqlColumn

    type: InferenceColumnType
    row_count: int
    distinct_row_count: int
    is_nullable: bool
    null_count: int
    min_value: Optional[T]
    max_value: Optional[T]

    @property
    def is_empty(self) -> bool:
        """Whether the column has any rows"""
        return self.row_count == 0


@dataclass(frozen=True)
class TableProperties:
    """Holds properties of a table and its columns that were extracted from the data warehouse."""

    column_props: InitVar[List[ColumnProperties]]

    table: SqlTable
    columns: Dict[SqlColumn, ColumnProperties] = field(default_factory=lambda: {}, init=False)

    def __post_init__(self, column_props: List[ColumnProperties]) -> None:  # noqa: D
        for col in column_props:
            self.columns[col.column] = col


@dataclass(frozen=True)
class DataWarehouseInferenceContext(InferenceContext):
    """The inference context for a data warehouse. Holds statistics and metadata about each table and column."""

    table_props: InitVar[List[TableProperties]]

    tables: Dict[SqlTable, TableProperties] = field(default_factory=lambda: {}, init=False)
    columns: Dict[SqlColumn, ColumnProperties] = field(default_factory=lambda: {}, init=False)

    def __post_init__(self, table_props: List[TableProperties]) -> None:  # noqa: D
        for stats in table_props:
            self.tables[stats.table] = stats
            for column in stats.columns.values():
                self.columns[column.column] = column


@contextlib.contextmanager
def _default_table_progress(table: SqlTable, index: int, total: int) -> Iterator[None]:
    yield


class DataWarehouseInferenceContextProvider(InferenceContextProvider[DataWarehouseInferenceContext], ABC):
    """Provides inference context from a data warehouse by querying data from its tables."""

    def __init__(self, client: SqlClient, tables: List[SqlTable], max_sample_size: int = 10000) -> None:
        """Initialize the class.

        client: the underlying SQL engine client that will be used for querying table data.
        tables: an exhaustive list of all tables that should be queried.
        max_sample_size: max number of rows to sample from each table
        """
        self._client = client
        self.tables = tables
        self.max_sample_size = max_sample_size

    @abstractmethod
    def _get_table_properties(self, table: SqlTable) -> TableProperties:
        """Fetch properties about a single table by querying the warehouse."""
        raise NotImplementedError

    def get_context(
        self,
        table_progress: Callable[[SqlTable, int, int], ContextManager[None]] = _default_table_progress,
    ) -> DataWarehouseInferenceContext:
        """Query the data warehouse for statistics about all tables and populate a context with it."""
        table_props_list: List[TableProperties] = []
        for i, table in enumerate(self.tables):
            with table_progress(table, i, len(self.tables)):
                table_props = self._get_table_properties(table)
                table_props_list.append(table_props)
        return DataWarehouseInferenceContext(table_props=table_props_list)



import logging
from typing import Sequence

import pytest

from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.node_evaluator import (
    NodeEvaluatorForLinkableInstances,
    LinkableInstanceSatisfiabilityEvaluation,
    JoinLinkableInstancesRecipe,
)
from metricflow.dataflow.builder.partitions import PartitionTimeDimensionJoinDescription
from metricflow.dataset.dataset import DataSet
from metricflow.model.semantic_model import SemanticModel
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.node_processor import PreDimensionJoinNodeProcessor
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.specs import (
    DimensionSpec,
    IdentifierSpec,
    LinklessIdentifierSpec,
    TimeDimensionSpec,
    LinkableInstanceSpec,
)
from metricflow.time.time_granularity import TimeGranularity
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository

logger = logging.getLogger(__name__)


@pytest.fixture
def node_evaluator(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_model: SemanticModel,
    dataflow_plan_builder: DataflowPlanBuilder[DataSourceDataSet],
    time_spine_source: TimeSpineSource,
) -> NodeEvaluatorForLinkableInstances:  # noqa: D
    """Return a node evaluator using the nodes in data_source_name_to_nodes"""
    node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver = DataflowPlanNodeOutputDataSetResolver(
        column_association_resolver=DefaultColumnAssociationResolver(simple_semantic_model),
        semantic_model=simple_semantic_model,
        time_spine_source=time_spine_source,
    )

    source_nodes = tuple(consistent_id_object_repository.simple_model_read_nodes.values())

    return NodeEvaluatorForLinkableInstances(
        data_source_semantics=simple_semantic_model.data_source_semantics,
        # Use all nodes in the simple model as candidates for joins.
        nodes_available_for_joins=source_nodes,
        node_data_set_resolver=node_data_set_resolver,
    )


def make_multihop_node_evaluator(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    multihop_semantic_model: SemanticModel,
    desired_linkable_specs: Sequence[LinkableInstanceSpec],
    time_spine_source: TimeSpineSource,
) -> NodeEvaluatorForLinkableInstances:  # noqa: D
    """Return a node evaluator using the nodes in multihop_data_source_name_to_nodes"""
    node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver = DataflowPlanNodeOutputDataSetResolver(
        column_association_resolver=DefaultColumnAssociationResolver(multihop_semantic_model),
        semantic_model=multihop_semantic_model,
        time_spine_source=time_spine_source,
    )

    node_processor = PreDimensionJoinNodeProcessor(
        data_source_semantics=multihop_semantic_model.data_source_semantics,
        node_data_set_resolver=node_data_set_resolver,
    )

    nodes_available_for_joins = node_processor.remove_unnecessary_nodes(
        desired_linkable_specs=desired_linkable_specs,
        nodes=consistent_id_object_repository.multihop_model_source_nodes,
        metric_time_dimension_reference=DataSet.metric_time_dimension_reference(),
    )

    nodes_available_for_joins = node_processor.add_multi_hop_joins(
        desired_linkable_specs=desired_linkable_specs, nodes=nodes_available_for_joins
    )

    return NodeEvaluatorForLinkableInstances(
        data_source_semantics=multihop_semantic_model.data_source_semantics,
        nodes_available_for_joins=nodes_available_for_joins,
        node_data_set_resolver=node_data_set_resolver,
    )


def test_node_evaluator_with_no_linkable_specs(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    bookings_source_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    evaluation = node_evaluator.evaluate_node(required_linkable_specs=[], start_node=bookings_source_node)
    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(), joinable_linkable_specs=(), join_recipes=(), unjoinable_linkable_specs=()
    )


def test_node_evaluator_with_unjoinable_specs(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    bookings_source_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="verification_type",
                identifier_links=(LinklessIdentifierSpec.from_element_name("verification"),),
            )
        ],
        start_node=bookings_source_node,
    )
    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(),
        join_recipes=(),
        unjoinable_linkable_specs=(
            DimensionSpec(
                element_name="verification_type",
                identifier_links=(LinklessIdentifierSpec.from_element_name("verification"),),
            ),
        ),
    )


def test_node_evaluator_with_local_spec(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Tests the case where the requested linkable spec in available in the start node."""
    bookings_source_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[DimensionSpec(element_name="is_instant", identifier_links=())],
        start_node=bookings_source_node,
    )
    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(DimensionSpec(element_name="is_instant", identifier_links=()),),
        joinable_linkable_specs=(),
        join_recipes=(),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_local_spec_using_primary_identifier(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Tests the case where the requested linkable spec with an identifier link is available in the start node."""
    bookings_source_node = consistent_id_object_repository.simple_model_read_nodes["users_latest"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="home_state_latest", identifier_links=(LinklessIdentifierSpec.from_element_name("user"),)
            )
        ],
        start_node=bookings_source_node,
    )

    assert evaluation == (
        LinkableInstanceSatisfiabilityEvaluation(
            local_linkable_specs=(
                DimensionSpec(
                    element_name="home_state_latest",
                    identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                ),
            ),
            joinable_linkable_specs=(),
            join_recipes=(),
            unjoinable_linkable_specs=(),
        )
    )


def test_node_evaluator_with_local_spec_using_primary_composite_identifier(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Similar to test_node_evaluator_with_local_spec_using_primary_identifier, but with a composite identifier"""
    bookings_source_node = consistent_id_object_repository.composite_model_read_nodes["users_source"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="country", identifier_links=(LinklessIdentifierSpec.from_element_name("user_team"),)
            )
        ],
        start_node=bookings_source_node,
    )

    assert (
        LinkableInstanceSatisfiabilityEvaluation(
            local_linkable_specs=(
                DimensionSpec(
                    element_name="country",
                    identifier_links=(LinklessIdentifierSpec(element_name="user_team", identifier_links=()),),
                ),
            ),
            joinable_linkable_specs=(),
            join_recipes=(),
            unjoinable_linkable_specs=(),
        )
        == evaluation
    )


def test_node_evaluator_with_joined_spec(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Tests the case where the requested linkable spec is available if another node is joined."""
    bookings_source_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(element_name="is_instant", identifier_links=()),
            DimensionSpec(
                element_name="country_latest",
                identifier_links=(LinklessIdentifierSpec.from_element_name(element_name="listing"),),
            ),
            DimensionSpec(
                element_name="capacity_latest",
                identifier_links=(LinklessIdentifierSpec.from_element_name(element_name="listing"),),
            ),
        ],
        start_node=bookings_source_node,
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(DimensionSpec(element_name="is_instant", identifier_links=()),),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="country_latest",
                identifier_links=(LinklessIdentifierSpec(element_name="listing", identifier_links=()),),
            ),
            DimensionSpec(
                element_name="capacity_latest",
                identifier_links=(LinklessIdentifierSpec(element_name="listing", identifier_links=()),),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.simple_model_read_nodes["listings_latest"],
                join_on_identifier=LinklessIdentifierSpec(element_name="listing", identifier_links=()),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="country_latest",
                        identifier_links=(LinklessIdentifierSpec(element_name="listing", identifier_links=()),),
                    ),
                    DimensionSpec(
                        element_name="capacity_latest",
                        identifier_links=(LinklessIdentifierSpec(element_name="listing", identifier_links=()),),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_joined_spec_on_unique_id(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Similar to test_node_evaluator_with_joined_spec() but using a unique identifier."""
    listings_node = consistent_id_object_repository.simple_model_read_nodes["listings_latest"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="company_name",
                identifier_links=(LinklessIdentifierSpec.from_element_name(element_name="user"),),
            ),
        ],
        start_node=listings_node,
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="company_name",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.simple_model_read_nodes["companies"],
                join_on_identifier=LinklessIdentifierSpec(element_name="user", identifier_links=()),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="company_name",
                        identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_multiple_joined_specs(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Tests the case where multiple nodes need to be joined to get all linkable specs."""
    views_source = consistent_id_object_repository.simple_model_read_nodes["views_source"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="home_state_latest",
                identifier_links=(LinklessIdentifierSpec.from_element_name(element_name="user"),),
            ),
            IdentifierSpec(
                element_name="user",
                identifier_links=(LinklessIdentifierSpec.from_element_name(element_name="listing"),),
            ),
        ],
        start_node=views_source,
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            IdentifierSpec(
                element_name="user",
                identifier_links=(LinklessIdentifierSpec(element_name="listing", identifier_links=()),),
            ),
            DimensionSpec(
                element_name="home_state_latest",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.simple_model_read_nodes["listings_latest"],
                join_on_identifier=LinklessIdentifierSpec(element_name="listing", identifier_links=()),
                satisfiable_linkable_specs=[
                    IdentifierSpec(
                        element_name="user",
                        identifier_links=(LinklessIdentifierSpec(element_name="listing", identifier_links=()),),
                    )
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            ),
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.simple_model_read_nodes["users_latest"],
                join_on_identifier=LinklessIdentifierSpec(element_name="user", identifier_links=()),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="home_state_latest",
                        identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                    )
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_multihop_joined_spec(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    multi_hop_join_semantic_model: SemanticModel,
    time_spine_source: TimeSpineSource,
) -> None:
    """Tests the case where multiple nodes need to be joined to get all linkable specs."""
    txn_source = consistent_id_object_repository.multihop_model_read_nodes["account_month_txns"]

    linkable_specs = [
        DimensionSpec(
            element_name="customer_name",
            identifier_links=(
                LinklessIdentifierSpec.from_element_name(element_name="account_id"),
                LinklessIdentifierSpec.from_element_name(element_name="customer_id"),
            ),
        ),
    ]

    multihop_node_evaluator = make_multihop_node_evaluator(
        consistent_id_object_repository=consistent_id_object_repository,
        multihop_semantic_model=multi_hop_join_semantic_model,
        desired_linkable_specs=linkable_specs,
        time_spine_source=time_spine_source,
    )

    evaluation = multihop_node_evaluator.evaluate_node(
        required_linkable_specs=linkable_specs,
        start_node=txn_source,
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="customer_name",
                identifier_links=(
                    LinklessIdentifierSpec.from_element_name(element_name="account_id"),
                    LinklessIdentifierSpec.from_element_name(element_name="customer_id"),
                ),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=evaluation.join_recipes[0].node_to_join,
                join_on_identifier=LinklessIdentifierSpec(element_name="account_id", identifier_links=()),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="customer_name",
                        identifier_links=(
                            LinklessIdentifierSpec(element_name="account_id", identifier_links=()),
                            LinklessIdentifierSpec(element_name="customer_id", identifier_links=()),
                        ),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(
                    PartitionTimeDimensionJoinDescription(
                        start_node_time_dimension_spec=TimeDimensionSpec(
                            element_name="ds_partitioned", identifier_links=(), time_granularity=TimeGranularity.DAY
                        ),
                        node_to_join_time_dimension_spec=TimeDimensionSpec(
                            element_name="ds_partitioned",
                            identifier_links=(),
                            time_granularity=TimeGranularity.DAY,
                        ),
                    ),
                ),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_partition_joined_spec(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Tests the case where the joined node required a partitioned join."""
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="home_state",
                identifier_links=(LinklessIdentifierSpec.from_element_name(element_name="user"),),
            ),
        ],
        start_node=consistent_id_object_repository.simple_model_read_nodes["id_verifications"],
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="home_state",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.simple_model_read_nodes["users_ds_source"],
                join_on_identifier=LinklessIdentifierSpec(element_name="user", identifier_links=()),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="home_state",
                        identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(
                    PartitionTimeDimensionJoinDescription(
                        start_node_time_dimension_spec=TimeDimensionSpec(
                            element_name="ds_partitioned",
                            identifier_links=(),
                        ),
                        node_to_join_time_dimension_spec=TimeDimensionSpec(
                            element_name="ds_partitioned",
                            identifier_links=(),
                        ),
                    ),
                ),
            ),
        ),
        unjoinable_linkable_specs=(),
    )



from abc import ABC
from typing import TypeVar

VisitorOutputT = TypeVar("VisitorOutputT")


class Visitable(ABC):
    """An object that follows the visitor pattern: https://en.wikipedia.org/wiki/Visitor_pattern

    Helps to perform processing on heterogeneous types in a structured way. This class doesn't do anything, but it's a
    place to centralize this comment :)

    If we can get something like the following to work, it would be ideal.

    def accept(visitor: VisitorT[VisitorOutputT]) -> VisitorOutputT.
       pass

    """

    pass



from metricflow.instances import InstanceSet
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.instance_converters import (
    CreateSelectColumnsWithMeasuresAggregated,
    FilterElements,
)
from metricflow.plan_conversion.select_column_gen import SelectColumnSet
from metricflow.sql.sql_exprs import (
    SqlFunction,
    SqlFunctionExpression,
)
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository

__SOURCE_TABLE_ALIAS = "a"


def __get_filtered_measure_instance_set(
    data_source_name: str, measure_name: str, object_repo: ConsistentIdObjectRepository
) -> InstanceSet:
    """Gets an InstanceSet consisting of only the measure instance matching the given name and data source"""
    dataset = object_repo.simple_model_data_sets[data_source_name]
    instance_set = dataset.instance_set
    include_specs = [
        instance.spec for instance in instance_set.measure_instances if instance.spec.element_name == measure_name
    ]
    return FilterElements(include_specs=include_specs).transform(instance_set)


def test_sum_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_model: SemanticModel,
) -> None:
    """Checks for function expression handling for booking_value, a SUM type metric in the simple model"""
    measure_name = "booking_value"
    instance_set = __get_filtered_measure_instance_set("bookings_source", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DefaultColumnAssociationResolver(simple_semantic_model),
        simple_semantic_model.data_source_semantics,
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlFunctionExpression)
    assert expr.sql_function == SqlFunction.SUM


def test_sum_boolean_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_model: SemanticModel,
) -> None:
    """Checks for function expression handling for instant_bookings, a SUM_BOOLEAN type metric in the simple model"""
    measure_name = "instant_bookings"
    instance_set = __get_filtered_measure_instance_set("bookings_source", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DefaultColumnAssociationResolver(simple_semantic_model),
        simple_semantic_model.data_source_semantics,
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlFunctionExpression)
    # The SUM_BOOLEAN aggregation type is transformed to SUM at model parsing time
    assert expr.sql_function == SqlFunction.SUM


def test_avg_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_model: SemanticModel,
) -> None:
    """Checks for function expression handling for average_booking_value, an AVG type metric in the simple model"""
    measure_name = "average_booking_value"
    instance_set = __get_filtered_measure_instance_set("bookings_source", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DefaultColumnAssociationResolver(simple_semantic_model),
        simple_semantic_model.data_source_semantics,
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlFunctionExpression)
    assert expr.sql_function == SqlFunction.AVERAGE


def test_count_distinct_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_model: SemanticModel,
) -> None:
    """Checks for function expression handling for bookers, a COUNT_DISTINCT type metric in the simple model"""
    measure_name = "bookers"
    instance_set = __get_filtered_measure_instance_set("bookings_source", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DefaultColumnAssociationResolver(simple_semantic_model),
        simple_semantic_model.data_source_semantics,
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlFunctionExpression)
    assert expr.sql_function == SqlFunction.COUNT_DISTINCT


def test_max_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_model: SemanticModel,
) -> None:
    """Checks for function expression handling for largest_listing, a MAX type metric in the simple model"""
    measure_name = "largest_listing"
    instance_set = __get_filtered_measure_instance_set("listings_latest", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DefaultColumnAssociationResolver(simple_semantic_model),
        simple_semantic_model.data_source_semantics,
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlFunctionExpression)
    assert expr.sql_function == SqlFunction.MAX


def test_min_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_model: SemanticModel,
) -> None:
    """Checks for function expression handling for smallest_listing, a MIN type metric in the simple model"""
    measure_name = "smallest_listing"
    instance_set = __get_filtered_measure_instance_set("listings_latest", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DefaultColumnAssociationResolver(simple_semantic_model),
        simple_semantic_model.data_source_semantics,
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlFunctionExpression)
    assert expr.sql_function == SqlFunction.MIN



from typing import List

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.materialization import Materialization
from metricflow.model.objects.metric import Metric
from metricflow.model.objects.base import HashableBaseModel


class UserConfiguredModel(HashableBaseModel):
    """Model holds all the information the SemanticLayer needs to render a query"""

    data_sources: List[DataSource]
    metrics: List[Metric]
    materializations: List[Materialization] = []



import pytest

from metricflow.model.objects.data_source import MutabilityType, Mutability
from metricflow.model.objects.elements.dimension import Dimension, DimensionType, DimensionTypeParams
from metricflow.model.validations.validator_helpers import ModelValidationException
from metricflow.test.model.validations.helpers import data_source_with_guaranteed_meta
from metricflow.time.time_granularity import TimeGranularity


@pytest.mark.skip("TODO: Will convert to validation rule")
def test_data_source_invalid_sql() -> None:  # noqa:D
    with pytest.raises(ModelValidationException, match=r"Invalid SQL"):
        data_source_with_guaranteed_meta(
            name="invalid_sql_source",
            sql_query="SELECT foo FROM bar;",
            dimensions=[
                Dimension(
                    name="ds",
                    type=DimensionType.TIME,
                    type_params=DimensionTypeParams(
                        time_granularity=TimeGranularity.DAY,
                    ),
                )
            ],
            mutability=Mutability(type=MutabilityType.IMMUTABLE),
        )



import collections

from dataclasses import dataclass
from typing import Sequence, Tuple

from metricflow.specs import MeasureSpec


@dataclass(frozen=True)
class GroupedMeasureSpecsByAdditiveness:
    """Results after grouping measures by their additive properties"""

    grouped_semi_additive_measures: Sequence[Tuple[MeasureSpec, ...]]
    additive_measures: Tuple[MeasureSpec, ...]


def group_measure_specs_by_additiveness(measure_specs: Sequence[MeasureSpec]) -> GroupedMeasureSpecsByAdditiveness:
    """Bucket the provided measure specs by

    - Additive Measures
    - Semi-additive measures containing the same non-additive dimension attributes
    """
    bucket = collections.defaultdict(list)
    additive_bucket = []
    for spec in measure_specs:
        non_additive_dimension = spec.non_additive_dimension
        if non_additive_dimension:
            bucket[non_additive_dimension.bucket_hash].append(spec)
        else:
            additive_bucket.append(spec)
    return GroupedMeasureSpecsByAdditiveness(
        grouped_semi_additive_measures=tuple(tuple(measures) for measures in bucket.values()),
        additive_measures=tuple(additive_bucket),
    )



"""Nodes used in defining an SQL query plan."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Optional, List, Generic, Sequence, Tuple

from metricflow.dag.mf_dag import DagNode, DisplayedProperty, MetricFlowDag, NodeId
from metricflow.dag.id_generation import (
    SQL_PLAN_SELECT_STATEMENT_ID_PREFIX,
    SQL_PLAN_TABLE_FROM_CLAUSE_ID_PREFIX,
)
from metricflow.dataflow.sql_table import SqlTable
from metricflow.sql.sql_exprs import SqlExpressionNode
from metricflow.visitor import VisitorOutputT

logger = logging.getLogger(__name__)


class SqlQueryPlanNode(DagNode, ABC):
    """Modeling a SQL query plan like a data flow plan as well.

    In that model:
    * A source node that takes data into the graph e.g. SQL tables or SQL queries in a literal string format.
    * SELECT queries can be modeled as nodes that transform the data.
    * Statements like ALTER TABLE don't fit well, but they could be modeled as just a single sink node.
    * SQL queries in where conditions could be modeled as another SqlQueryPlan.
    * SqlRenderableNode() indicates nodes where plan generation can begin. Generally, this will be all nodes except
      the SqlTableFromClauseNode() since my_table.my_column wouldn't be a valid SQL query.

    Is there an existing library that can do this?
    """

    def __init__(self, node_id: NodeId, parent_nodes: Sequence[SqlQueryPlanNode]) -> None:  # noqa: D:
        self._parent_nodes = parent_nodes
        super().__init__(node_id=node_id)

    @property
    def parent_nodes(self) -> List[SqlQueryPlanNode]:  # noqa: D
        return list(self._parent_nodes)

    @abstractmethod
    def accept(self, visitor: SqlQueryPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        """Called when a visitor needs to visit this node."""
        pass

    @property
    @abstractmethod
    def is_table(self) -> bool:
        """Returns whether this node resolves to a table (vs. a query)"""
        pass

    @property
    @abstractmethod
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:
        """If possible, return this as a select statement node."""
        pass


class SqlQueryPlanNodeVisitor(Generic[VisitorOutputT], ABC):
    """An object that can be used to visit the nodes of an SQL query.

    See similar visitor DataflowPlanVisitor.
    """

    @abstractmethod
    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_table_from_clause_node(self, node: SqlTableFromClauseNode) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> VisitorOutputT:  # noqa: D
        pass


@dataclass(frozen=True)
class SqlSelectColumn:
    """Represents a column in the select clause of an SQL query."""

    expr: SqlExpressionNode
    # Always require a column alias for simplicity.
    column_alias: str


class SqlJoinType(Enum):
    """Enumerates the different kinds of SQL joins.

    The value is the SQL string to be used when rendering the join.
    """

    LEFT_OUTER = "LEFT OUTER JOIN"
    FULL_OUTER = "FULL OUTER JOIN"
    INNER = "INNER JOIN"

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}.{self.name}"


@dataclass(frozen=True)
class SqlJoinDescription:
    """Describes how sources should be joined together."""

    # The source that goes on the right side of the JOIN keyword.
    right_source: SqlQueryPlanNode
    right_source_alias: str
    on_condition: SqlExpressionNode
    join_type: SqlJoinType


@dataclass(frozen=True)
class SqlOrderByDescription:  # noqa: D
    expr: SqlExpressionNode
    desc: bool


class SqlSelectStatementNode(SqlQueryPlanNode):
    """Represents an SQL Select statement."""

    def __init__(  # noqa: D
        self,
        description: str,
        select_columns: Tuple[SqlSelectColumn, ...],
        from_source: SqlQueryPlanNode,
        from_source_alias: str,
        joins_descs: Tuple[SqlJoinDescription, ...],
        group_bys: Tuple[SqlSelectColumn, ...],
        order_bys: Tuple[SqlOrderByDescription, ...],
        where: Optional[SqlExpressionNode] = None,
        limit: Optional[int] = None,
    ) -> None:
        self._description = description
        assert select_columns
        self._select_columns = select_columns
        # Sources that belong in a from clause. CTEs could be captured in a separate field.
        self._from_source = from_source
        self._from_source_alias = from_source_alias
        self._join_descs = joins_descs
        self._group_bys = group_bys
        self._where = where
        self._order_bys = order_bys

        if limit is not None:
            assert limit >= 0
        self._limit = limit

        super().__init__(
            node_id=self.create_unique_id(),
            parent_nodes=[self._from_source] + [x.right_source for x in self._join_descs],
        )

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_PLAN_SELECT_STATEMENT_ID_PREFIX

    @property
    def description(self) -> str:  # noqa: D
        return self._description

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:  # noqa: D
        return (
            super().displayed_properties
            + [DisplayedProperty(f"col{i}", column) for i, column in enumerate(self._select_columns)]
            + [DisplayedProperty("from_source", self.from_source)]
            + [DisplayedProperty(f"join_{i}", join_desc) for i, join_desc in enumerate(self._join_descs)]
            + [DisplayedProperty(f"group_by{i}", group_by) for i, group_by in enumerate(self._group_bys)]
            + [DisplayedProperty("where", self._where)]
            + [DisplayedProperty(f"order_by{i}", order_by) for i, order_by in enumerate(self._order_bys)]
        )

    @property
    def select_columns(self) -> Tuple[SqlSelectColumn, ...]:  # noqa: D
        return self._select_columns

    @property
    def from_source(self) -> SqlQueryPlanNode:  # noqa: D
        return self._from_source

    @property
    def from_source_alias(self) -> str:  # noqa: D
        return self._from_source_alias

    @property
    def join_descs(self) -> Tuple[SqlJoinDescription, ...]:  # noqa: D
        return self._join_descs

    @property
    def group_bys(self) -> Tuple[SqlSelectColumn, ...]:  # noqa: D
        return self._group_bys

    @property
    def where(self) -> Optional[SqlExpressionNode]:  # noqa: D
        return self._where

    @property
    def order_bys(self) -> Tuple[SqlOrderByDescription, ...]:  # noqa: D
        return self._order_bys

    def accept(self, visitor: SqlQueryPlanNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_select_statement_node(self)

    @property
    def is_table(self) -> bool:  # noqa: D
        return False

    @property
    def limit(self) -> Optional[int]:  # noqa: D
        return self._limit

    @property
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:  # noqa: D
        return self


class SqlTableFromClauseNode(SqlQueryPlanNode):
    """An SQL table that can go in the FROM clause."""

    def __init__(self, sql_table: SqlTable) -> None:  # noqa: D
        self._sql_table = sql_table
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_PLAN_TABLE_FROM_CLAUSE_ID_PREFIX

    @property
    def description(self) -> str:  # noqa: D
        return f"Read from {self._sql_table.sql}"

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:  # noqa: D
        return super().displayed_properties + [
            DisplayedProperty("table_id", self._sql_table.sql),
        ]

    def accept(self, visitor: SqlQueryPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D
        return visitor.visit_table_from_clause_node(self)

    @property
    def sql_table(self) -> SqlTable:  # noqa: D
        return self._sql_table

    @property
    def is_table(self) -> bool:  # noqa: D
        return True

    @property
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:  # noqa: D
        return None


class SqlSelectQueryFromClauseNode(SqlQueryPlanNode):
    """An SQL select query that can go in the FROM clause."""

    def __init__(self, select_query: str) -> None:  # noqa: D
        self._select_query = select_query
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_PLAN_TABLE_FROM_CLAUSE_ID_PREFIX

    @property
    def description(self) -> str:  # noqa: D
        return "Read From a Select Query"

    def accept(self, visitor: SqlQueryPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D
        return visitor.visit_query_from_clause_node(self)

    @property
    def select_query(self) -> str:  # noqa: D
        return self._select_query

    @property
    def is_table(self) -> bool:  # noqa: D
        return False

    @property
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:  # noqa: D
        return None


class SqlQueryPlan(MetricFlowDag[SqlQueryPlanNode]):  # noqa: D
    """Model for an SQL Query as a DAG."""

    def __init__(self, plan_id: str, render_node: SqlQueryPlanNode) -> None:
        """Constructor.

        Args:
            plan_id: The ID to associate with this plan.
            render_node: The node from which to start rendering the SQL query.
        """
        self._render_node = render_node
        super().__init__(dag_id=plan_id, sink_nodes=[self._render_node])

    @property
    def render_node(self) -> SqlQueryPlanNode:  # noqa: D
        return self._render_node



from metricflow.model.parsing.config_linter import ConfigLinter


def test_lint_dir(config_linter_model_path: str) -> None:  # noqa: D
    issues = ConfigLinter().lint_dir(dir_path=config_linter_model_path)

    assert len(issues.all_issues) == 1
    assert len(issues.errors) == 1
    assert 'duplication of key "data_source"' in issues.errors[0].as_readable_str()



import logging

import pytest
from typing import Generator

from metricflow.sql_clients.sql_utils import make_sql_client
from metricflow.protocols.sql_client import SqlClient
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState

logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def sql_client(mf_test_session_state: MetricFlowTestSessionState) -> Generator[SqlClient, None, None]:
    """Provides a base SqlClient for use in integration tests requiring warehouse access."""
    sql_client = make_sql_client(
        url=mf_test_session_state.sql_engine_url,
        password=mf_test_session_state.sql_engine_password,
    )
    logger.info(f"Creating schema '{mf_test_session_state.mf_system_schema}'")
    sql_client.create_schema(mf_test_session_state.mf_system_schema)
    if mf_test_session_state.mf_system_schema != mf_test_session_state.mf_source_schema:
        logger.info(f"Creating schema '{mf_test_session_state.mf_source_schema}'")
        sql_client.create_schema(mf_test_session_state.mf_source_schema)

    yield sql_client

    logger.info(f"Dropping schema '{mf_test_session_state.mf_system_schema}'")
    sql_client.drop_schema(mf_test_session_state.mf_system_schema, cascade=True)
    if mf_test_session_state.mf_system_schema != mf_test_session_state.mf_source_schema:
        logger.info(f"Dropping schema '{mf_test_session_state.mf_source_schema}'")
        sql_client.drop_schema(mf_test_session_state.mf_source_schema, cascade=True)

    sql_client.close()
    return None



"""This module aims to help figure out the answer to the following scenario:

I have a node containing a measure that's needed for a metric, and I also have a list of dimensions that I need for that
measure.

Can I join that node with other nodes using a common identifier to get those dimensions? If so, what are those nodes,
what identifier should I join on, and if I do that join, what dimensions can be retrieved?

Note: the term "dimension" is used below, but it actually refers to any LinkableInstance. Also, when you see the term
"start_node", think "node containing the measure". Using the term "start_node" as the scenario of joining measure nodes
to dimension nodes is an easy case to explain, but there are non-measure nodes that can be joined with dimension nodes
to realize other planned features.
"""

from __future__ import annotations

import itertools
import logging
from dataclasses import dataclass
from typing import List, Sequence, Generic, TypeVar, Tuple

from metricflow.model.objects.elements.identifier import IdentifierType
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.partitions import PartitionJoinResolver
from metricflow.dataflow.dataflow_plan import (
    JoinDescription,
    BaseOutput,
    PartitionDimensionJoinDescription,
    PartitionTimeDimensionJoinDescription,
)
from metricflow.instances import InstanceSet
from metricflow.object_utils import pformat_big_objects
from metricflow.plan_conversion.sql_dataset import SqlDataSet
from metricflow.specs import (
    LinkableInstanceSpec,
    LinklessIdentifierSpec,
    InstanceSpecSet,
)
from metricflow.model.semantics.semantic_containers import DataSourceSemantics

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class JoinLinkableInstancesRecipe:
    """A recipe for how to join a node containing linkable instances to the "start_node".

    node_to_join contains the linkable instances that are needed - it should be filtered so that the output data set of
    that node only includes the identifier instance for the join and the instances associated with
    satisfiable_linkable_specs.
    """

    node_to_join: BaseOutput
    # The identifier to join "node_to_join" on.
    join_on_identifier: LinklessIdentifierSpec
    # The linkable instances from the query that can be satisfied if we join this node. Note that this is different from
    # the linkable specs in the node that can help to satisfy the query. e.g. "user_id__country" might be one of the
    # "satisfiable_linkable_specs", but "country" is the linkable spec in the node.
    satisfiable_linkable_specs: List[LinkableInstanceSpec]

    # The partitions to join on, if there are matching partitions between the start_node and node_to_join.
    join_on_partition_dimensions: Tuple[PartitionDimensionJoinDescription, ...]
    join_on_partition_time_dimensions: Tuple[PartitionTimeDimensionJoinDescription, ...]

    @property
    def join_description(self) -> JoinDescription:
        """The recipe as a join description to use in the dataflow plan node."""
        return JoinDescription(
            join_node=self.node_to_join,
            join_on_identifier=self.join_on_identifier,
            join_on_partition_dimensions=self.join_on_partition_dimensions,
            join_on_partition_time_dimensions=self.join_on_partition_time_dimensions,
        )


@dataclass(frozen=True)
class LinkableInstanceSatisfiabilityEvaluation:
    """Evaluation for a node on the ability to get linkable instances that we need to resolve a query.

    Includes whether the node contains the linkable instances that we need (local_linkable_specs), or if it can have
    them joined in (joinable_linkable_specs). Linkable instances that can't be satisfied either locally or via joins are
    listed in unjoinable_linkable_specs.
    """

    local_linkable_specs: Tuple[LinkableInstanceSpec, ...]
    joinable_linkable_specs: Tuple[LinkableInstanceSpec, ...]
    join_recipes: Tuple[JoinLinkableInstancesRecipe, ...]
    unjoinable_linkable_specs: Tuple[LinkableInstanceSpec, ...]


SourceDataSetT = TypeVar("SourceDataSetT", bound=SqlDataSet)


class NodeEvaluatorForLinkableInstances(Generic[SourceDataSetT]):
    """Helps to evaluate if linkable instances can be obtained using the given node, with joins if necessary.

    For example, consider a "start_node" containing the "bookings" measure, "is_instant" dimension, and "listing_id"
    identifier with nodes_available_for_joins including a node with the "listing_id" identifier, and the "country"
    dimension.

    We want to know if we can get "bookings", "is_instant", "listing_id__country" using the start_node. The result
    should be that we know: "is_instant" is available locally (i.e. in the same node), and if we join another node
    containing "listing_id" and "country" by "listing_id", we can get "listing_id__country".
    """

    def __init__(
        self,
        data_source_semantics: DataSourceSemantics,
        nodes_available_for_joins: Sequence[BaseOutput[SourceDataSetT]],
        node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver[SourceDataSetT],
    ) -> None:
        """Constructor

        Args:
            data_source_semantics: Needed to resolve partition dimensions.
            nodes_available_for_joins: Nodes that contain linkable instances and may be joined with the "start_node"
            (e.g. the node containing a desired measure) to retrieve the needed linkable instances.
            node_data_set_resolver: Figures out what data set is output by a node.
        """
        self._data_source_semantics = data_source_semantics
        self._nodes_available_for_joins = nodes_available_for_joins
        self._node_data_set_resolver = node_data_set_resolver
        self._partition_resolver = PartitionJoinResolver(self._data_source_semantics)

    def _find_joinable_candidate_nodes_that_can_satisfy_linkable_specs(
        self,
        start_node_spec_set: InstanceSpecSet,
        needed_linkable_specs: List[LinkableInstanceSpec],
    ) -> List[JoinLinkableInstancesRecipe]:
        """Get nodes that can be joined to get 1 or more of the "needed_linkable_specs".

        The returned list is ordered by the number of "needed_linkable_specs" that it can satisfy.
        """
        candidates_for_join: List[JoinLinkableInstancesRecipe] = []

        for node in self._nodes_available_for_joins:
            data_set: SqlDataSet = self._node_data_set_resolver.get_output_data_set(node)
            linkable_specs_in_node = data_set.instance_set.spec_set.linkable_specs
            identifier_specs_in_node = data_set.instance_set.spec_set.identifier_specs

            # For each unlinked identifier in the data set, create a candidate for joining.
            # For a data set to be useful for satisfying a linkable spec, it needs to have the identifier
            # and the linkable spec without the identifier. This allows joining based on the identifier, which will
            # then produce the linkable spec. See comments further below for more details.

            for identifier_spec_in_node in identifier_specs_in_node:
                # If an identifier has links, what that means and whether it can be used is unclear at the moment,
                # so skip it.
                if len(identifier_spec_in_node.identifier_links) > 0:
                    continue

                identifier_instance = None
                for instance in data_set.instance_set.identifier_instances:
                    if instance.spec == identifier_spec_in_node:
                        identifier_instance = instance
                        break

                if identifier_instance is None:
                    raise RuntimeError(f"Could not find identifier instance with name ({identifier_spec_in_node})")

                assert (
                    len(identifier_instance.defined_from) == 1
                ), f"Did not get exactly 1 defined_from in {identifier_instance}"

                ident = self._data_source_semantics.get_identifier_in_data_source(identifier_instance.defined_from[0])
                if ident is None:
                    raise RuntimeError(f"Invalid DataSourceElementReference {identifier_instance.defined_from[0]}")
                if ident.type not in (IdentifierType.PRIMARY, IdentifierType.UNIQUE):
                    continue

                linkless_identifier_spec_in_node = LinklessIdentifierSpec.from_element_name(
                    identifier_spec_in_node.element_name
                )

                satisfiable_linkable_specs = []
                for needed_linkable_spec in needed_linkable_specs:
                    assert (
                        len(needed_linkable_spec.identifier_links) != 0
                    ), f"Invalid needed linkable spec passed in {needed_linkable_spec}"

                    # If the identifier in the data set matches the link, then it can be used for joins. For example,
                    # if the node has the identifier "user_id", and dimension "country" then it can be used for
                    # satisfying "user_id__country".
                    #
                    # Multi-hop example:
                    # required_linkable_spec = "user_id__device_id__platform"
                    # identifier_spec_in_data_set = "user_id"
                    #
                    # Then the data set must contain "device_id__platform", which is realized with
                    #
                    # required_linkable_spec.remove_first_identifier_link()
                    #
                    # We might also need to check the identifier type and see if it's the type of join we're allowing,
                    # but since we're doing all left joins now, it's been left out.

                    required_identifier_matches_data_set_identifier = (
                        needed_linkable_spec.identifier_links[0] == linkless_identifier_spec_in_node
                    )
                    needed_linkable_spec_in_node = (
                        needed_linkable_spec.without_first_identifier_link() in linkable_specs_in_node
                    )
                    if required_identifier_matches_data_set_identifier and needed_linkable_spec_in_node:
                        satisfiable_linkable_specs.append(needed_linkable_spec)

                # If this node can satisfy some linkable specs, it could be useful to join on, so add it to the
                # candidate list.
                if len(satisfiable_linkable_specs) > 0:
                    join_on_partition_dimensions = self._partition_resolver.resolve_partition_dimension_joins(
                        start_node_spec_set=start_node_spec_set,
                        node_to_join_spec_set=data_set.instance_set.spec_set,
                    )
                    join_on_partition_time_dimensions = self._partition_resolver.resolve_partition_time_dimension_joins(
                        start_node_spec_set=start_node_spec_set,
                        node_to_join_spec_set=data_set.instance_set.spec_set,
                    )
                    candidates_for_join.append(
                        JoinLinkableInstancesRecipe(
                            node_to_join=node,
                            join_on_identifier=linkless_identifier_spec_in_node,
                            satisfiable_linkable_specs=satisfiable_linkable_specs,
                            join_on_partition_dimensions=join_on_partition_dimensions,
                            join_on_partition_time_dimensions=join_on_partition_time_dimensions,
                        )
                    )

        # Return with the candidate set that can satisfy the most linkable specs at the front.
        return sorted(
            candidates_for_join,
            key=lambda x: len(x.satisfiable_linkable_specs),
            reverse=True,
        )

    @staticmethod
    def _update_candidates_that_can_satisfy_linkable_specs(
        candidates_for_join: List[JoinLinkableInstancesRecipe],
        already_satisfisfied_linkable_specs: List[LinkableInstanceSpec],
    ) -> List[JoinLinkableInstancesRecipe]:
        """Update / filter candidates_for_join based on linkable instance specs that we have already satisfied.

        Those linkable instances are no longer needed because those were satisfied previously by joining a node
        containing some needed linkable instances to the "start_node". This method will then remove the satisfied
        linkable instances from the candidate recipes, and if a candidate recipe doesn't have any linkable instances
        that can help satisfy the query, it is removed.
        """
        updated_candidate_data_sets: List[JoinLinkableInstancesRecipe] = []
        for candidate_for_join in candidates_for_join:
            updated_satisfiable_linkable_specs = list(
                set(candidate_for_join.satisfiable_linkable_specs) - set(already_satisfisfied_linkable_specs)
            )

            if len(updated_satisfiable_linkable_specs) > 0:
                updated_candidate_data_sets.append(
                    JoinLinkableInstancesRecipe(
                        node_to_join=candidate_for_join.node_to_join,
                        join_on_identifier=candidate_for_join.join_on_identifier,
                        satisfiable_linkable_specs=updated_satisfiable_linkable_specs,
                        join_on_partition_dimensions=candidate_for_join.join_on_partition_dimensions,
                        join_on_partition_time_dimensions=candidate_for_join.join_on_partition_time_dimensions,
                    )
                )
        return sorted(
            updated_candidate_data_sets,
            key=lambda x: len(x.satisfiable_linkable_specs),
            reverse=True,
        )

    def evaluate_node(
        self,
        start_node: BaseOutput[SourceDataSetT],
        required_linkable_specs: List[LinkableInstanceSpec],
    ) -> LinkableInstanceSatisfiabilityEvaluation:
        """Evaluates if the "required_linkable_specs" can be realized by joining the "start_node" with other nodes.

        In other words, given the data set associated with "start_node":

        * Can all "required_linkable_specs" be retrieved from the start_node? (These would be considered "local").
        * If not, can they be retrieved by joining an available node though a common identifier?
        * If so, return all possible ways (by joining different nodes) that can be done.
        """
        candidate_instance_set: InstanceSet = self._node_data_set_resolver.get_output_data_set(start_node).instance_set
        candidate_spec_set = candidate_instance_set.spec_set

        logger.debug(f"Candidate spec set is:\n{pformat_big_objects(candidate_spec_set)}")

        data_set_linkable_specs = candidate_spec_set.linkable_specs

        # These are linkable specs in the same data set as the measure. Those are considered "local".
        local_linkable_specs = []

        # These are linkable specs that aren't in the data set, but they might be able to be joined in.
        possibly_joinable_linkable_specs = []

        # Group required_linkable_specs into local / un-joinable / or possibly joinable.
        unjoinable_linkable_specs = []
        for required_linkable_spec in required_linkable_specs:
            is_local = required_linkable_spec in data_set_linkable_specs
            is_unjoinable = (
                len(required_linkable_spec.identifier_links) == 0
                or required_linkable_spec.identifier_links[0] not in data_set_linkable_specs
            )
            if is_local:
                local_linkable_specs.append(required_linkable_spec)
            elif is_unjoinable:
                unjoinable_linkable_specs.append(required_linkable_spec)
            else:
                possibly_joinable_linkable_specs.append(required_linkable_spec)

        candidates_for_join = self._find_joinable_candidate_nodes_that_can_satisfy_linkable_specs(
            start_node_spec_set=candidate_spec_set, needed_linkable_specs=possibly_joinable_linkable_specs
        )
        join_candidates: List[JoinLinkableInstancesRecipe] = []

        logger.info("Looping over nodes that can be joined to get the required linkable specs")

        # Using a greedy approach, try to get the "possibly_joinable_linkable_specs" by iteratively joining nodes with
        # the most matching linkable specs. We try to join nodes with the most matching specs to minimize the number of
        # joins that we have to do to. A knapsack solution is ideal, but punting on that for simplicity.
        while len(possibly_joinable_linkable_specs) > 0:
            logger.info(f"Looking for linkable specs:\n{pformat_big_objects(possibly_joinable_linkable_specs)}")

            # We've run out of candidate data sets, but there are more linkable specs that we need. That means the
            # rest of the linkable specs can't be joined in, and we're left with unjoinable specs remaining.
            if len(candidates_for_join) == 0:
                logger.info(
                    "There are no more candidate nodes that can be joined, but not all linkable specs have "
                    "been acquired."
                )
                unjoinable_linkable_specs.extend(possibly_joinable_linkable_specs)
                break

            # Join the best candidate to realize the linkable specs
            next_candidate = candidates_for_join.pop(0)
            logger.info(f"The next candidate node to be joined is:\n{pformat_big_objects(next_candidate)}")
            join_candidates.append(next_candidate)

            # Update the candidates. Since we'll be joined/ing the previously selected candidate, we no longer need
            # to get those linkable specs from the remaining candidates, and we also need to re-rank based on the new
            # requirements.
            candidates_for_join = self._update_candidates_that_can_satisfy_linkable_specs(
                candidates_for_join=candidates_for_join,
                already_satisfisfied_linkable_specs=next_candidate.satisfiable_linkable_specs,
            )

            # The once possibly joinable specs are definitely joinable and no longer need to be searched for.
            # Remove from "possibly_joinable_linkable_specs"
            possibly_joinable_linkable_specs = [
                x for x in possibly_joinable_linkable_specs if x not in next_candidate.satisfiable_linkable_specs
            ]

        logger.info("Done evaluating possible joins")
        return LinkableInstanceSatisfiabilityEvaluation(
            local_linkable_specs=tuple(local_linkable_specs),
            joinable_linkable_specs=tuple(
                itertools.chain.from_iterable([x.satisfiable_linkable_specs for x in join_candidates])
            ),
            join_recipes=tuple(join_candidates),
            unjoinable_linkable_specs=tuple(unjoinable_linkable_specs),
        )



"""Functions to help generate a text representation of a dataflow plan."""

import logging
import textwrap

import jinja2

from metricflow.dag.mf_dag import DagNodeVisitor, DagNode
from metricflow.object_utils import pformat_big_objects

logger = logging.getLogger(__name__)


class MetricFlowDagToText(DagNodeVisitor[str]):
    """Converts the given node and parents (recursively) to a text representation.

    The text representation should be similar to the XML format:

    <FilterElementsNode>
      ...
    </>
    """

    # Parameters for controlling the text output.
    MAX_WIDTH = 60

    def _format_to_text(self, node: DagNode, inner_contents: str) -> str:
        """Convert the given node to the text representation.

        Properties is a dict from the property name to the property value that should be printed for the node.
        """
        # Generate the descriptions for the node
        node_fields = []
        for displayed_property in node.displayed_properties:

            if len(str(displayed_property.value)) > self.MAX_WIDTH or "\n" in str(displayed_property.value):
                value_str_split = pformat_big_objects(displayed_property.value).split("\n")
                max_value_str_length = max([len(x) for x in value_str_split])
                node_fields.append(
                    jinja2.Template(
                        textwrap.dedent(
                            """\
                            <!-- {{ key }} = {{ padding }} -->
                            """
                        ),
                        undefined=jinja2.StrictUndefined,
                    ).render(
                        key=displayed_property.key,
                        padding=" "
                        * (
                            (len("<!--   ") + max_value_str_length + len(" -->"))
                            - len("<!-- ")
                            - len(displayed_property.key)
                            - len(" = ")
                            - len("-->")
                        ),
                    )
                )
                for value_str in value_str_split:
                    node_fields.append(
                        jinja2.Template(
                            textwrap.dedent(
                                """\
                                <!--   {{ value }} {{ padding }} -->
                                """
                            ),
                            undefined=jinja2.StrictUndefined,
                        ).render(
                            value=value_str,
                            padding=" "
                            * (
                                (len("<!-- ") + max_value_str_length + len(" -->"))
                                - len("<!-- ")
                                - len(value_str)
                                - len(" -->")
                            ),
                        )
                    )
            else:
                node_fields.append(
                    jinja2.Template(
                        textwrap.dedent(
                            """\
                            <!-- {{ key }} = {{ value }} -->
                            """
                        ),
                        undefined=jinja2.StrictUndefined,
                    ).render(
                        key=displayed_property.key,
                        value=displayed_property.value,
                    )
                )

        return jinja2.Template(
            textwrap.dedent(
                """\
                <{{ node_class }}{%- if not inner_contents and not node_fields %}/>{%- else %}>
                    {%- if node_fields %}
                    {{ node_fields | indent(4) }}
                    {%- endif %}
                    {%- if inner_contents %}
                    {{ inner_contents | indent(4) }}
                    {%- endif %}
                </{{ node_class }}>
                {%- endif %}
                """
            ),
            undefined=jinja2.StrictUndefined,
        ).render(
            node_class=node.__class__.__name__,
            node_fields="\n".join(node_fields),
            inner_contents=inner_contents,
        )

    def visit_node(self, node: DagNode) -> str:  # noqa: D
        parent_node_descriptions = []
        for parent_node in node.parent_nodes:
            parent_node_descriptions.append(parent_node.accept_dag_node_visitor(self))

        return self._format_to_text(node=node, inner_contents="\n".join(parent_node_descriptions))

    def to_text(self, root_node: DagNode) -> str:
        """Converts the dag starting from the given root node to a text representation.

        The text representation is similar to XML.
        """
        return root_node.accept_dag_node_visitor(self)



import logging

from metricflow.model.objects.metric import Metric, MetricInputMeasure, MetricType, MetricTypeParams
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.transformations.transform_rule import ModelTransformRule

logger = logging.getLogger(__name__)


class ModelTransformError(Exception):  # noqa: D
    pass


class CreateProxyMeasureRule(ModelTransformRule):
    """Adds a proxy metric for measures that have the create_metric flag set, if it does not already exist.

    Also checks that a defined metric with the same name as a measure is a proxy metric.
    """

    @staticmethod
    def transform_model(model: UserConfiguredModel) -> UserConfiguredModel:
        """Creates measure proxy metrics for measures with `create_metric==True`"""
        for data_source in model.data_sources:
            for measure in data_source.measures:
                if not measure.create_metric:
                    continue

                add_metric = True
                for metric in model.metrics:
                    if metric.name == measure.name:
                        if metric.type != MetricType.MEASURE_PROXY:
                            raise ModelTransformError(
                                f"Cannot have metric with the same name as a measure ({measure.name}) that is not a "
                                f"proxy for that measure"
                            )
                        logger.warning(
                            f"Metric already exists with name ({measure.name}). *Not* adding measure proxy metric for "
                            f"that measure"
                        )
                        add_metric = False

                if add_metric is True:
                    model.metrics.append(
                        Metric(
                            name=measure.name,
                            type=MetricType.MEASURE_PROXY,
                            type_params=MetricTypeParams(
                                measures=[MetricInputMeasure(name=measure.name)],
                                expr=measure.name,
                            ),
                        )
                    )

        return model



from collections import defaultdict
import os
from pathlib import Path
from typing import Dict, Literal, Union, List, TypedDict
from typing_extensions import NotRequired

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from metricflow.dataflow.sql_table import SqlTable
from metricflow.inference.models import InferenceResult, InferenceSignalType
from metricflow.inference.renderer.base import InferenceRenderer

yaml = YAML()


class RenderedTimeColumnConfigTypeParams(TypedDict):  # noqa: D
    time_granularity: Literal["day"]
    is_primary: NotRequired[bool]


class RenderedColumnConfig(TypedDict):  # noqa: D
    name: str
    type: str
    type_params: NotRequired[RenderedTimeColumnConfigTypeParams]


class ConfigFileRenderer(InferenceRenderer):
    """Writes inference results to a set of config files"""

    UNKNOWN_FIELD_VALUE = "FIXME"

    def __init__(self, dir_path: Union[str, Path], overwrite: bool = False) -> None:
        """Initializes the renderer.

        dir_path: The path to the config directory
        overwrite: If set to False, will raise error if the directory exists
        """
        dir_path = os.path.abspath(dir_path)

        if not overwrite and os.path.exists(dir_path):
            raise ValueError("ConfigFileRender.overwrite is False but path exists.")

        if os.path.isfile(dir_path):
            raise ValueError("ConfigFileRenderer `dir_path` is a file.")

        self.dir_path = dir_path

    def _get_filename_for_table(self, table: SqlTable) -> str:
        return os.path.abspath(os.path.join(self.dir_path, f"{table.sql}.yaml"))

    def _fixme(self, comment: str) -> str:
        return f"FIXME: {comment}"

    def _render_id_columns(self, results: List[InferenceResult]) -> List[CommentedMap]:
        type_map = {
            InferenceSignalType.ID.PRIMARY: "primary",
            InferenceSignalType.ID.FOREIGN: "foreign",
            InferenceSignalType.ID.UNIQUE: "unique",
        }

        rendered: List[RenderedColumnConfig] = [
            {
                "name": result.column.column_name,
                "type": type_map.get(result.type_node, ConfigFileRenderer.UNKNOWN_FIELD_VALUE),
            }
            for result in results
            if result.type_node.is_subtype_of(InferenceSignalType.ID.UNKNOWN)
        ]

        return rendered

    def _render_dimension_columns(self, results: List[InferenceResult]) -> List[CommentedMap]:
        type_map = {
            InferenceSignalType.DIMENSION.TIME: "time",
            InferenceSignalType.DIMENSION.PRIMARY_TIME: "time",
            InferenceSignalType.DIMENSION.CATEGORICAL: "categorical",
        }

        rendered: List[CommentedMap] = []
        for result in results:
            if not result.type_node.is_subtype_of(InferenceSignalType.DIMENSION.UNKNOWN):
                continue

            result_data: CommentedMap = CommentedMap(
                {
                    "name": result.column.column_name,
                    "type": type_map.get(result.type_node, ConfigFileRenderer.UNKNOWN_FIELD_VALUE),
                }
            )

            if result_data["type"] == ConfigFileRenderer.UNKNOWN_FIELD_VALUE:
                result_data.yaml_add_eol_comment(self._fixme("unknown field value"), "type")

            if result.type_node.is_subtype_of(InferenceSignalType.DIMENSION.TIME):
                type_params: CommentedMap = {"time_granularity": "day"}
                if result.type_node.is_subtype_of(InferenceSignalType.DIMENSION.PRIMARY_TIME):
                    type_params["is_primary"] = True
                result_data["type_params"] = type_params

            if len(result.problems) > 0:
                result_data.yaml_set_comment_before_after_key(
                    key="name",
                    before=f"{ConfigFileRenderer.UNKNOWN_FIELD_VALUE}: " + ", ".join(result.problems),
                )

            rendered.append(result_data)

        return rendered

    def _get_comments_for_unknown_columns(self, results: List[InferenceResult]) -> List[str]:
        delim = "\n  - "
        return [
            self._fixme(result.column.column_name + delim + delim.join(result.problems))
            for result in results
            if result.type_node == InferenceSignalType.UNKNOWN
        ]

    def _get_configuration_data_for_table(self, table: SqlTable, results: List[InferenceResult]) -> Dict:
        data = CommentedMap(
            {
                "data_source": {
                    "name": table.table_name,
                    "sql_table": table.sql,
                    "identifiers": self._render_id_columns(results),
                    "dimensions": self._render_dimension_columns(results),
                    "measures": [],
                }
            }
        )

        header_comments = [self._fixme("Unreviewed inferred config file")] + self._get_comments_for_unknown_columns(
            results
        )
        data.yaml_set_comment_before_after_key(
            key="data_source",
            before="\n".join(header_comments),
        )

        return data

    def render(self, results: List[InferenceResult]) -> None:
        """Render the inference results to files in the configured directory."""
        os.makedirs(self.dir_path, exist_ok=True)

        results_by_table: Dict[SqlTable, List[InferenceResult]] = defaultdict(list)
        for result in results:
            results_by_table[result.column.table].append(result)

        for table, results in results_by_table.items():
            table_data = self._get_configuration_data_for_table(table, results)
            table_path = self._get_filename_for_table(table)
            with open(table_path, "w") as table_file:
                yaml.dump(table_data, table_file)



from __future__ import annotations

import pytest

from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer, DefaultSqlQueryPlanRenderer


@pytest.fixture
def default_sql_plan_renderer() -> SqlQueryPlanRenderer:  # noqa: D
    return DefaultSqlQueryPlanRenderer()



import pytest

from metricflow.model.model_validator import ModelValidator
from metricflow.model.objects.data_source import Mutability, MutabilityType
from metricflow.model.objects.elements.dimension import Dimension, DimensionType, DimensionTypeParams
from metricflow.model.objects.elements.identifier import Identifier, IdentifierType
from metricflow.model.objects.elements.measure import Measure, AggregationType
from metricflow.model.objects.metric import MetricType, MetricTypeParams
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import ModelValidationException
from metricflow.references import DimensionReference, IdentifierReference, TimeDimensionReference
from metricflow.test.fixtures.table_fixtures import DEFAULT_DS
from metricflow.test.model.validations.helpers import data_source_with_guaranteed_meta, metric_with_guaranteed_meta
from metricflow.time.time_granularity import TimeGranularity


@pytest.mark.skip("TODO: Will convert to validation rule")
def test_metric_missing_measure() -> None:  # noqa:D
    with pytest.raises(ModelValidationException):
        measure_name = "my_measure"
        measure2_name = "nonexistent_measure"
        model_validator = ModelValidator()
        model_validator.checked_validations(
            UserConfiguredModel(
                data_sources=[
                    data_source_with_guaranteed_meta(
                        name="sum_measure",
                        sql_query="SELECT foo FROM bar",
                        measures=[Measure(name=measure_name, agg=AggregationType.SUM)],
                        mutability=Mutability(type=MutabilityType.IMMUTABLE),
                    )
                ],
                metrics=[
                    metric_with_guaranteed_meta(
                        name="metric_with_nonexistent_measure",
                        type=MetricType.MEASURE_PROXY,
                        type_params=MetricTypeParams(measures=[measure2_name]),
                    )
                ],
            )
        )


def test_metric_no_time_dim_dim_only_source() -> None:  # noqa:D
    dim_name = "country"
    dim2_name = "ename"
    measure_name = "foo"
    model_validator = ModelValidator()
    model_validator.checked_validations(
        UserConfiguredModel(
            data_sources=[
                data_source_with_guaranteed_meta(
                    name="sum_measure",
                    sql_query="SELECT foo, country FROM bar",
                    measures=[],
                    dimensions=[Dimension(name=dim_name, type=DimensionType.CATEGORICAL)],
                    mutability=Mutability(type=MutabilityType.IMMUTABLE),
                ),
                data_source_with_guaranteed_meta(
                    name="sum_measure2",
                    sql_query="SELECT foo, country FROM bar",
                    measures=[
                        Measure(
                            name=measure_name,
                            agg=AggregationType.SUM,
                            agg_time_dimension=dim2_name,
                        )
                    ],
                    dimensions=[
                        Dimension(name=dim_name, type=DimensionType.CATEGORICAL),
                        Dimension(
                            name=dim2_name,
                            type=DimensionType.TIME,
                            type_params=DimensionTypeParams(
                                is_primary=True,
                                time_granularity=TimeGranularity.DAY,
                            ),
                        ),
                    ],
                    mutability=Mutability(type=MutabilityType.IMMUTABLE),
                ),
            ],
            metrics=[
                metric_with_guaranteed_meta(
                    name="metric_with_no_time_dim",
                    type=MetricType.MEASURE_PROXY,
                    type_params=MetricTypeParams(measures=[measure_name]),
                )
            ],
            materializations=[],
        )
    )


def test_metric_no_time_dim() -> None:  # noqa:D
    with pytest.raises(ModelValidationException):
        dim_name = "country"
        measure_name = "foo"
        model_validator = ModelValidator()
        model_validator.checked_validations(
            UserConfiguredModel(
                data_sources=[
                    data_source_with_guaranteed_meta(
                        name="sum_measure",
                        sql_query="SELECT foo, country FROM bar",
                        measures=[Measure(name=measure_name, agg=AggregationType.SUM)],
                        dimensions=[
                            Dimension(
                                name=dim_name,
                                type=DimensionType.CATEGORICAL,
                            )
                        ],
                        mutability=Mutability(type=MutabilityType.IMMUTABLE),
                    )
                ],
                metrics=[
                    metric_with_guaranteed_meta(
                        name="metric_with_no_time_dim",
                        type=MetricType.MEASURE_PROXY,
                        type_params=MetricTypeParams(measures=[measure_name]),
                    )
                ],
                materializations=[],
            )
        )


def test_metric_multiple_primary_time_dims() -> None:  # noqa:D
    with pytest.raises(ModelValidationException):
        dim_name = "date_created"
        dim2_name = "date_deleted"
        measure_name = "foo"
        model_validator = ModelValidator()
        model_validator.checked_validations(
            UserConfiguredModel(
                data_sources=[
                    data_source_with_guaranteed_meta(
                        name="sum_measure",
                        sql_query="SELECT foo, date_created, date_deleted FROM bar",
                        measures=[Measure(name=measure_name, agg=AggregationType.SUM)],
                        dimensions=[
                            Dimension(
                                name=dim_name,
                                type=DimensionType.TIME,
                                type_params=DimensionTypeParams(
                                    time_granularity=TimeGranularity.DAY,
                                ),
                            ),
                            Dimension(
                                name=dim2_name,
                                type=DimensionType.TIME,
                                type_params=DimensionTypeParams(
                                    time_granularity=TimeGranularity.DAY,
                                ),
                            ),
                        ],
                        mutability=Mutability(type=MutabilityType.IMMUTABLE),
                    )
                ],
                metrics=[
                    metric_with_guaranteed_meta(
                        name="foo",
                        type=MetricType.MEASURE_PROXY,
                        type_params=MetricTypeParams(measures=[measure_name]),
                    )
                ],
                materializations=[],
            )
        )


def test_generated_metrics_only() -> None:  # noqa:D
    dim_reference = DimensionReference(element_name="dim")

    dim2_reference = TimeDimensionReference(element_name=DEFAULT_DS)
    measure_name = "measure"
    identifier_reference = IdentifierReference(element_name="primary")
    data_source = data_source_with_guaranteed_meta(
        name="dim1",
        sql_query=f"SELECT {dim_reference.element_name}, {measure_name} FROM bar",
        measures=[Measure(name=measure_name, agg=AggregationType.SUM, agg_time_dimension=dim2_reference.element_name)],
        dimensions=[
            Dimension(name=dim_reference.element_name, type=DimensionType.CATEGORICAL),
            Dimension(
                name=dim2_reference.element_name,
                type=DimensionType.TIME,
                type_params=DimensionTypeParams(
                    is_primary=True,
                    time_granularity=TimeGranularity.DAY,
                ),
            ),
        ],
        mutability=Mutability(type=MutabilityType.IMMUTABLE),
        identifiers=[
            Identifier(name=identifier_reference.element_name, type=IdentifierType.PRIMARY),
        ],
    )
    data_source.measures[0].create_metric = True

    ModelValidator().checked_validations(
        UserConfiguredModel(
            data_sources=[data_source],
            metrics=[],
            materializations=[],
        )
    )



from _pytest.fixtures import FixtureRequest

from metricflow.dataflow.sql_table import SqlTable
from metricflow.engine.metricflow_engine import MetricFlowQueryRequest
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.integration.conftest import IntegrationTestHelpers
from metricflow.test.plan_utils import make_schema_replacement_function, assert_snapshot_text_equal

# Needed as the table alias can vary from run to run.
_EXCLUDE_TABLE_ALIAS_REGEX = "^.*_src.*$"


def test_render_query(  # noqa: D
    request: FixtureRequest, mf_test_session_state: MetricFlowTestSessionState, it_helpers: IntegrationTestHelpers
) -> None:
    result = it_helpers.mf_engine.explain(
        MetricFlowQueryRequest.create_with_random_request_id(
            metric_names=["bookings"],
            group_by_names=["ds"],
        )
    )

    assert_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        group_id=result.__class__.__name__,
        snapshot_id="query0",
        snapshot_text=result.rendered_sql.sql_query,
        snapshot_file_extension=".sql",
        incomparable_strings_replacement_function=make_schema_replacement_function(
            system_schema=it_helpers.mf_system_schema, source_schema=it_helpers.source_schema
        ),
        exclude_line_regex=_EXCLUDE_TABLE_ALIAS_REGEX,
        additional_sub_directories_for_snapshots=(it_helpers.sql_client.__class__.__name__,),
    )


def test_render_write_to_table_query(  # noqa: D
    request: FixtureRequest, mf_test_session_state: MetricFlowTestSessionState, it_helpers: IntegrationTestHelpers
) -> None:
    output_table = SqlTable(schema_name=it_helpers.mf_system_schema, table_name="test_table")

    result = it_helpers.mf_engine.explain(
        MetricFlowQueryRequest.create_with_random_request_id(
            metric_names=["bookings"], group_by_names=["ds"], output_table=output_table.sql
        )
    )

    assert_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        group_id=result.__class__.__name__,
        snapshot_id="query0",
        snapshot_text=result.rendered_sql.sql_query,
        snapshot_file_extension=".sql",
        incomparable_strings_replacement_function=make_schema_replacement_function(
            system_schema=it_helpers.mf_system_schema, source_schema=it_helpers.source_schema
        ),
        exclude_line_regex=_EXCLUDE_TABLE_ALIAS_REGEX,
        additional_sub_directories_for_snapshots=(it_helpers.sql_client.__class__.__name__,),
    )



import textwrap

from metricflow.model.objects.common import YamlConfigFile
from metricflow.model.objects.constraints.where import WhereClauseConstraint
from metricflow.model.objects.metric import CumulativeMetricWindow, MetricType, MetricInputMeasure
from metricflow.model.parsing.dir_to_model import parse_yaml_files_to_model
from metricflow.model.validations.validator_helpers import ModelValidationException
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.time.time_granularity import TimeGranularity


def test_legacy_measure_metric_parsing() -> None:
    """Test for parsing a simple metric specification with the `measure` parameter instead of `measures`"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: legacy_test
          type: measure_proxy
          type_params:
            measure: legacy_measure
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.metrics) == 1
    metric = build_result.model.metrics[0]
    assert metric.name == "legacy_test"
    assert metric.type is MetricType.MEASURE_PROXY
    assert metric.type_params.measure == MetricInputMeasure(name="legacy_measure")
    assert metric.type_params.measures is None


def test_legacy_metric_input_measure_object_parsing() -> None:
    """Test for parsing a simple metric specification with the `measure` parameter set with object notation"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: legacy_test
          type: measure_proxy
          type_params:
            measure:
              name: legacy_measure_from_object
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.metrics) == 1
    metric = build_result.model.metrics[0]
    assert metric.type_params.measure == MetricInputMeasure(name="legacy_measure_from_object")


def test_metric_metadata_parsing() -> None:
    """Test for asserting that internal metadata is parsed into the Metric object"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: metadata_test
          type: measure_proxy
          type_params:
            measures:
              - metadata_test_measure
        """
    )
    file = YamlConfigFile(filepath="test_dir/inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.metrics) == 1
    metric = build_result.model.metrics[0]
    assert metric.metadata is not None
    assert metric.metadata.repo_file_path == "test_dir/inline_for_test"
    assert metric.metadata.file_slice.filename == "inline_for_test"
    expected_metadata_content = textwrap.dedent(
        """\
        name: metadata_test
        type: measure_proxy
        type_params:
          measures:
          - metadata_test_measure
        """
    )
    assert metric.metadata.file_slice.content == expected_metadata_content


def test_ratio_metric_parsing() -> None:
    """Test for parsing a ratio metric specification with numerator and denominator"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: ratio_test
          type: ratio
          type_params:
            numerator: numerator_measure
            denominator: denominator_measure
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.metrics) == 1
    metric = build_result.model.metrics[0]
    assert metric.name == "ratio_test"
    assert metric.type is MetricType.RATIO
    assert metric.type_params.numerator == MetricInputMeasure(name="numerator_measure")
    assert metric.type_params.denominator == MetricInputMeasure(name="denominator_measure")
    assert metric.type_params.measures is None


def test_ratio_metric_input_measure_object_parsing() -> None:
    """Test for parsing a ratio metric specification with object inputs for numerator and denominator"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: ratio_test
          type: ratio
          type_params:
            numerator:
              name: numerator_measure_from_object
            denominator:
              name: denominator_measure_from_object
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.metrics) == 1
    metric = build_result.model.metrics[0]
    assert metric.type_params.numerator == MetricInputMeasure(name="numerator_measure_from_object")
    assert metric.type_params.denominator == MetricInputMeasure(name="denominator_measure_from_object")


def test_expr_metric_parsing() -> None:
    """Test for parsing a metric specification with an expr and a list of measures"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: expr_test
          type: expr
          type_params:
            measures:
              - measure_one
              - measure_two
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.metrics) == 1
    metric = build_result.model.metrics[0]
    assert metric.name == "expr_test"
    assert metric.type is MetricType.EXPR
    assert metric.type_params.measures == [
        MetricInputMeasure(name="measure_one"),
        MetricInputMeasure(name="measure_two"),
    ]


def test_expr_metric_input_measure_object_parsing() -> None:
    """Test for parsing a metric specification with object inputs for the list of measures"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: expr_test
          type: expr
          type_params:
            measures:
              - name: measure_one_from_object
              - name: measure_two_from_object
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.metrics) == 1
    metric = build_result.model.metrics[0]
    assert metric.name == "expr_test"
    assert metric.type is MetricType.EXPR
    assert metric.type_params.measures == [
        MetricInputMeasure(name="measure_one_from_object"),
        MetricInputMeasure(name="measure_two_from_object"),
    ]


def test_cumulative_window_metric_parsing() -> None:
    """Test for parsing a metric specification with a cumulative window"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: cumulative_test
          type: cumulative
          type_params:
            measures:
              - cumulative_measure
            window: "7 days"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.metrics) == 1
    metric = build_result.model.metrics[0]
    assert metric.name == "cumulative_test"
    assert metric.type is MetricType.CUMULATIVE
    assert metric.type_params.measures == [MetricInputMeasure(name="cumulative_measure")]
    assert metric.type_params.window == CumulativeMetricWindow(count=7, granularity=TimeGranularity.DAY)


def test_grain_to_date_metric_parsing() -> None:
    """Test for parsing a metric specification with the grain to date cumulative setting"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: grain_to_date_test
          type: cumulative
          type_params:
            measures:
              - cumulative_measure
            grain_to_date: "week"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.metrics) == 1
    metric = build_result.model.metrics[0]
    assert metric.name == "grain_to_date_test"
    assert metric.type is MetricType.CUMULATIVE
    assert metric.type_params.measures == [MetricInputMeasure(name="cumulative_measure")]
    assert metric.type_params.window is None
    assert metric.type_params.grain_to_date is TimeGranularity.WEEK


def test_constraint_metric_parsing() -> None:
    """Test for parsing a metric specification with a constraint included"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: constraint_test
          type: measure_proxy
          type_params:
            measures:
              - input_measure
          constraint: "some_dimension IN ('value1', 'value2')"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.metrics) == 1
    metric = build_result.model.metrics[0]
    assert metric.name == "constraint_test"
    assert metric.type is MetricType.MEASURE_PROXY
    assert metric.constraint == WhereClauseConstraint(
        where="some_dimension IN ('value1', 'value2')",
        linkable_names=["some_dimension"],
        sql_params=SqlBindParameters(),
    )


def test_invalid_metric_type_parsing_error() -> None:
    """Test for error detection when parsing a metric specification with an invalid MetricType input value"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: invalid_type_test
          type: this is not a valid type
          type_params:
            measures:
              - input_measure
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])
    assert build_result.issues.has_blocking_issues
    assert "'this is not a valid type' is not one of" in str(ModelValidationException(build_result.issues.all_issues))


def test_invalid_cumulative_metric_window_format_parsing_error() -> None:
    """Test for errror detection when parsing malformed cumulative metric window entries"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: invalid_cumulative_format_test
          type: cumulative
          type_params:
            measures:
              - cumulative_measure
            window: "7 days long"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])
    assert build_result.issues.has_blocking_issues
    assert "Invalid window" in str(ModelValidationException(build_result.issues.all_issues))


def test_invalid_cumulative_metric_window_granularity_parsing_error() -> None:
    """Test for errror detection when parsing malformed cumulative metric window entries"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: invalid_cumulative_granularity_test
          type: cumulative
          type_params:
            measures:
              - cumulative_measure
            window: "7 moons"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])
    assert build_result.issues.has_blocking_issues
    assert "Invalid time granularity" in str(ModelValidationException(build_result.issues.all_issues))


def test_invalid_cumulative_metric_window_count_parsing_error() -> None:
    """Test for errror detection when parsing malformed cumulative metric window entries"""
    yaml_contents = textwrap.dedent(
        """\
        metric:
          name: invalid_cumulative_count_test
          type: cumulative
          type_params:
            measures:
              - cumulative_measure
            window: "six days"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])
    assert build_result.issues.has_blocking_issues
    assert "Invalid count" in str(ModelValidationException(build_result.issues.all_issues))



import pytest

from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter
from metricflow.plan_conversion.time_spine import TimeSpineSource


@pytest.fixture(scope="session")
def dataflow_to_sql_converter(  # noqa: D
    simple_semantic_model: SemanticModel,
    time_spine_source: TimeSpineSource,
) -> DataflowToSqlQueryPlanConverter[DataSourceDataSet]:
    return DataflowToSqlQueryPlanConverter[DataSourceDataSet](
        column_association_resolver=DefaultColumnAssociationResolver(simple_semantic_model),
        semantic_model=simple_semantic_model,
        time_spine_source=time_spine_source,
    )



import textwrap

import jinja2

from metricflow.dag.dag_to_text import MetricFlowDagToText
from metricflow.execution.execution_plan import ExecutionPlan


def execution_plan_to_text(execution_plan: ExecutionPlan) -> str:
    """Convert the execution plan to a text form that's like XML"""
    if len(execution_plan.sink_nodes) != 1:
        raise RuntimeError("Currently, only 1 sink node is supported in the execution plan.")

    node = execution_plan.sink_nodes[0]

    return jinja2.Template(
        textwrap.dedent(
            """\
            <{{ node_class }}{%- if not inner_contents and not node_fields %}/>{%- else %}>
                {%- if inner_contents %}
                {{ inner_contents | indent(4) }}
                {%- endif %}
            </{{ node_class }}>
            {%- endif %}
            """
        ),
        undefined=jinja2.StrictUndefined,
    ).render(
        node_class=execution_plan.__class__.__name__,
        inner_contents=MetricFlowDagToText().to_text(node),
    )



from metricflow.dataflow.sql_column import SqlColumn
from metricflow.dataflow.sql_table import SqlTable
from metricflow.inference.context.data_warehouse import (
    ColumnProperties,
    DataWarehouseInferenceContext,
    InferenceColumnType,
    TableProperties,
)
from metricflow.inference.models import InferenceSignalType
import metricflow.inference.rule.defaults as defaults


def get_column_properties(column_str: str, type: InferenceColumnType, unique: bool) -> ColumnProperties:  # noqa: D
    return ColumnProperties(
        column=SqlColumn.from_string(column_str),
        type=type,
        row_count=10000,
        distinct_row_count=10000 if unique else 9000,
        is_nullable=False,
        null_count=0,
        min_value=0,
        max_value=9999,
    )


def test_any_identifier_by_name_matcher() -> None:  # noqa: D
    assert defaults.AnyIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.id", InferenceColumnType.INTEGER, True)
    )
    assert defaults.AnyIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.tableid", InferenceColumnType.INTEGER, True)
    )
    assert defaults.AnyIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.table_id", InferenceColumnType.INTEGER, True)
    )
    assert defaults.AnyIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.othertable_id", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.AnyIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.whatever", InferenceColumnType.INTEGER, True)
    )


def test_primary_identifier_by_name_matcher() -> None:  # noqa: D
    assert defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.id", InferenceColumnType.INTEGER, True)
    )
    assert defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.tableid", InferenceColumnType.INTEGER, True)
    )
    assert defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.table_id", InferenceColumnType.INTEGER, True)
    )
    assert defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.tables.table_id", InferenceColumnType.INTEGER, True)
    )
    assert defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.tables.tableid", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.othertable_id", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.othertableid", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.whatever", InferenceColumnType.INTEGER, True)
    )


def test_unique_identifier_by_distinct_count_matcher() -> None:  # noqa: D
    assert defaults.UniqueIdentifierByDistinctCountRule().match_column(
        get_column_properties("db.schema.table.unique_id", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.UniqueIdentifierByDistinctCountRule().match_column(
        get_column_properties("db.schema.table.unique_id", InferenceColumnType.STRING, False)
    )


def test_time_dimension_by_time_type_matcher() -> None:  # noqa: D
    assert defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.DATETIME, True)
    )

    assert not defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.FLOAT, True)
    )
    assert not defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.BOOLEAN, True)
    )
    assert not defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.STRING, True)
    )
    assert not defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.UNKNOWN, True)
    )


def test_primary_time_dimension_by_name_matcher() -> None:  # noqa: D
    assert defaults.PrimaryTimeDimensionByNameRule().match_column(
        get_column_properties("db.schema.table.ds", InferenceColumnType.DATETIME, True)
    )
    assert defaults.PrimaryTimeDimensionByNameRule().match_column(
        get_column_properties("db.schema.table.created_at", InferenceColumnType.DATETIME, True)
    )
    assert not defaults.PrimaryTimeDimensionByNameRule().match_column(
        get_column_properties("db.schema.table.bla", InferenceColumnType.DATETIME, True)
    )
    assert not defaults.PrimaryTimeDimensionByNameRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.DATETIME, True)
    )


def test_primary_time_dimension_if_only_time_rule() -> None:  # noqa: D
    table = SqlTable.from_string("db.schema.table")
    single_time_col_warehouse = DataWarehouseInferenceContext(
        table_props=[
            TableProperties(
                table=table,
                column_props=[
                    get_column_properties("db.schema.table.id", InferenceColumnType.INTEGER, True),
                    get_column_properties("db.schema.table.time", InferenceColumnType.DATETIME, True),
                ],
            )
        ]
    )
    single_time_col_signals = defaults.PrimaryTimeDimensionIfOnlyTimeRule().process(single_time_col_warehouse)
    assert len(single_time_col_signals) == 1
    assert single_time_col_signals[0].column == SqlColumn.from_string("db.schema.table.time")
    assert single_time_col_signals[0].type_node == InferenceSignalType.DIMENSION.PRIMARY_TIME

    many_time_col_warehouse = DataWarehouseInferenceContext(
        table_props=[
            TableProperties(
                table=table,
                column_props=[
                    get_column_properties("db.schema.table.id", InferenceColumnType.INTEGER, True),
                    get_column_properties("db.schema.table.time", InferenceColumnType.DATETIME, True),
                    get_column_properties("db.schema.table.othertime", InferenceColumnType.DATETIME, True),
                ],
            )
        ]
    )
    many_time_col_signals = defaults.PrimaryTimeDimensionIfOnlyTimeRule().process(many_time_col_warehouse)
    assert len(many_time_col_signals) == 0


def test_categorical_dimension_by_boolean_type_matcher() -> None:  # noqa: D
    assert defaults.CategoricalDimensionByBooleanTypeRule().match_column(
        get_column_properties("db.schema.table.dim", InferenceColumnType.BOOLEAN, True)
    )
    assert not defaults.CategoricalDimensionByBooleanTypeRule().match_column(
        get_column_properties("db.schema.table.bla", InferenceColumnType.FLOAT, True)
    )


def test_categorical_dimension_by_string_type_matcher() -> None:  # noqa: D
    assert defaults.CategoricalDimensionByStringTypeRule().match_column(
        get_column_properties("db.schema.table.dim", InferenceColumnType.STRING, True)
    )
    assert not defaults.CategoricalDimensionByStringTypeRule().match_column(
        get_column_properties("db.schema.table.bla", InferenceColumnType.FLOAT, True)
    )


def test_categorical_dimension_by_integer_type_matcher() -> None:  # noqa: D
    assert defaults.CategoricalDimensionByIntegerTypeRule().match_column(
        get_column_properties("db.schema.table.dim", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.CategoricalDimensionByIntegerTypeRule().match_column(
        get_column_properties("db.schema.table.bla", InferenceColumnType.FLOAT, True)
    )


def test_measure_by_real_type_matcher() -> None:  # noqa: D
    assert defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.FLOAT, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.DATETIME, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.BOOLEAN, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.STRING, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.UNKNOWN, True)
    )


def test_measure_by_integer_type_matcher() -> None:  # noqa: D
    assert defaults.MeasureByIntegerTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.BOOLEAN, True)
    )



import os
from typing import Optional


class EnvironmentVariable:
    """Represents an environment variable and provides a way to retrieve the associated value."""

    def __init__(self, name: str, default_value: str = None):
        """Constructor.

        Args:
            name: The name of the environment variable that this represents.
            default_value: The default value to return if this environment variable is not set.
        """
        self._name = name
        self._default_value = default_value

    def exists(self) -> bool:
        """Returns true if the this environment variable exists / is set."""
        return os.getenv(self._name) is not None

    def get(self) -> str:
        """Returns the value of this environment variable as a string.

        Returns the default value if not defined. If the environment variable is not defined and no default was
        specified, this throws an exception.
        """
        value = os.getenv(self._name)
        if value is None and self._default_value is not None:
            return self._default_value
        if value is None:
            raise RuntimeError(f"Environment variable {self._name} is not defined.")
        return value

    def get_optional(self) -> Optional[str]:
        """Returns value of env variable if it exists or None if it doesn't."""
        return os.getenv(self._name)

    def get_int(self) -> int:
        """Returns value of this environment variable as an integer.

        Returns the default value if not defined. If the environment variable is not defined and no default was
        specified, this throws an exception.
        """
        value = os.environ.get(self._name)
        if value is None and self._default_value is not None:
            try:
                return int(self._default_value)
            except ValueError as e:
                raise ValueError(
                    f"Default value for {self._name} is {self._default_value}, which is not an int."
                ) from e
        if value is None:
            raise RuntimeError(f"Environment variable {self._name} is not defined.")
        try:
            return int(value)
        except ValueError as e:
            raise ValueError(f"Environment variable {self._name} is {self._default_value}, which is not an int.") from e

    @staticmethod
    def _is_valid_bool(bool_str: str) -> bool:
        """Checks if the given string can be converted into a bool."""
        return bool_str.lower() == "true" or bool_str.lower() == "false"

    def get_bool(self) -> bool:
        """Returns the value of this environment variable as a bool.

        Returns the default value if not defined. If the environment variable is not defined and no default was
        specified, this throws an exception.
        """
        value = os.environ.get(self._name)

        if value is None:
            if self._default_value is None:
                raise RuntimeError(f"Environment variable {self._name} is not defined and has no default.")

            if not EnvironmentVariable._is_valid_bool(self._default_value):
                raise ValueError(
                    f"Default value for {self._name} is {self._default_value}, which is not true or false."
                )

            return self._default_value.lower() == "true"

        if not EnvironmentVariable._is_valid_bool(value):
            raise ValueError(f"Value for {self._name} is {value}, which is not true or false.")

        return value.lower() == "true"

    @property
    def name(self) -> str:  # noqa: D
        return self._name



from abc import ABC
from typing import Union

from metricflow.model.objects.base import FrozenBaseModel


class ColumnCorrelationKey(ABC, FrozenBaseModel):
    """Interface for a key object that is used to correlate columns between instance sets."""

    pass


class SingleColumnCorrelationKey(ColumnCorrelationKey):
    """Key to use when there's only 1 column association in an instance."""

    pass


class CompositeColumnCorrelationKey(ColumnCorrelationKey):
    """Key to use when there are multiple column associations in an instance"""

    sub_identifier: str


class ColumnAssociation(FrozenBaseModel):
    """Used to describe how an instance is associated with columns in table / SQL query.

    Generally there will be a 1:1 mapping, but for composite identifiers, it can map to multiple columns. For that case,
    this can be subclassed to add more context.
    """

    column_name: str
    # When an instance is passed from one dataflow node to another, we need to know how the columns from the input
    # corresponds to the columns from the output. Equality of this key is used to determine that relationship.
    # This could be made to be in a dictionary instead, but having it here means that it doesn't need to be hashable.
    column_correlation_key: Union[CompositeColumnCorrelationKey, SingleColumnCorrelationKey]



"""Tests for end to end object metadata parsing from the simple model YAML files"""

import os
from typing import Sequence

import pytest

from metricflow.model.objects.common import Metadata
from metricflow.model.objects.elements.measure import Measure
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.parsing.yaml_loader import YamlConfigLoader
from metricflow.model.semantic_model import SemanticModel


def test_data_source_metadata_parsing(simple_user_configured_model: UserConfiguredModel) -> None:
    """Tests internal metadata object parsing from a file into the Data Source model object

    This only tests some basic file name parsing for each data source since they are not guaranteed
    to be collected in the same file in the simple model, and the output here has been transformed
    so the YAML contents might or might not match.
    """
    assert len(simple_user_configured_model.data_sources) > 0
    for data_source in simple_user_configured_model.data_sources:
        assert (
            data_source.metadata is not None
        ), f"Metadata should always be parsed out of the model, but None found for data source: {data_source}!"
        _assert_metadata_filename_is_valid(data_source.metadata)


def test_metric_metadata_parsing(simple_user_configured_model: UserConfiguredModel) -> None:
    """Tests internal metadata object parsing from a file into the Metric model object

    This only tests some basic file name parsing for each metric since they are not guaranteed
    to be collected in the same file in the simple model, and the output here has been transformed
    so the YAML contents might or might not match.
    """
    assert len(simple_user_configured_model.metrics) > 0
    for metric in simple_user_configured_model.metrics:
        assert (
            metric.metadata is not None
        ), f"Metadata should always be parsed out of the model, but None found for metric: {metric}!"
        _assert_metadata_filename_is_valid(metric.metadata)


@pytest.mark.skip("TODO: Determine what to do with measure proxy metric metadata")
def test_metric_metadata_parsing_with_measure_proxy(multi_hop_join_semantic_model: SemanticModel) -> None:
    """Tests internal metadata object parsing from a file into the Metric model object via measure proxy

    The simple model has a broader array of metric definitions, but it does not appear to have a measure proxy
    added via transformation. This test includes such a metric.

    This only tests some basic file name parsing for each metric since they are not guaranteed
    to be collected in the same file in the simple model, and the output here has been transformed
    so the YAML contents might or might not match.
    """
    assert len(multi_hop_join_semantic_model.user_configured_model.metrics) > 0
    for metric in multi_hop_join_semantic_model.user_configured_model.metrics:
        assert (
            metric.metadata is not None
        ), f"Metadata should always be parsed out of the model, but None found for metric: {metric}!"
        _assert_metadata_filename_is_valid(metric.metadata)


def test_materialization_metadata_parsing(simple_user_configured_model: UserConfiguredModel) -> None:
    """Tests internal metadata object parsing from a file into the Materialization model object

    This only tests some basic file name parsing for each materialization since they are not guaranteed
    to be collected in the same file in the simple model, and the output here has been transformed
    so the YAML contents might or might not match.
    """
    assert len(simple_user_configured_model.materializations) > 0
    for materialization in simple_user_configured_model.materializations:
        assert (
            materialization.metadata is not None
        ), f"Metadata should always be parsed out of the model, but None found for materialization: {materialization}!"
        _assert_metadata_filename_is_valid(materialization.metadata)


def test_measure_metadata_parsing(simple_user_configured_model: UserConfiguredModel) -> None:
    """Tests internal metadata object parsing from a file into the Measure model object

    This tests the complete parsing process for Measure object metadata, including some baseline testing of things
    like file line number extraction and propagation. As with other cases, no assertions are made on the
    YAML contents themselves since they may change from the raw files into the UserConfiguredModel object we access
    here.
    """
    assert len(simple_user_configured_model.data_sources) > 0
    for data_source in simple_user_configured_model.data_sources:
        _assert_measure_metadata_is_valid(data_source.measures)


def _assert_metadata_filename_is_valid(metadata: Metadata) -> None:
    """Sequence of assertion steps to ensure the metadata object has consistent file name parsing"""
    assert YamlConfigLoader.is_valid_yaml_file_ending(metadata.repo_file_path), (
        f"Expected repo file path in measure metadata to be a yaml file with an appropriate ending. "
        f"Metadata: {metadata}"
    )

    assert (
        os.path.basename(metadata.repo_file_path) == metadata.file_slice.filename
    ), f"File name should be the final part of the repo file path. Metadata: {metadata}"
    assert (
        os.path.dirname(metadata.repo_file_path) != ""
    ), f"Expected repo file path to be fully resolved, but it is filename only. Metadata: {metadata}"


def _assert_measure_metadata_is_valid(measures: Sequence[Measure]) -> None:
    """Sequence of assertion steps to show that we are parsing metadata consistently for measures

    The assertions check that:
    1. Metadata is always set by the parser
    2. Metadata always contains a reasonable repo file path and file name
    3. Start and end line numbers are always increasing for every measure in the data source

    Since this test is operating on a transformed model, we do not make assertions about the raw YAML contents and how
    they relate to the properties of the measures themselves.
    """
    last_end_number = 0
    for measure in measures:
        assert (
            measure.metadata is not None
        ), f"Metadata should always be parsed out of the model, but None found for measure {measure}!"

        _assert_metadata_filename_is_valid(measure.metadata)

        assert measure.metadata.file_slice.start_line_number > last_end_number, (
            f"Start line numbers should always follow the end line number from the previous measure! "
            f"Metadata: {measure.metadata}"
        )
        assert measure.metadata.file_slice.end_line_number > measure.metadata.file_slice.start_line_number, (
            f"End line numbers should always be larger than start line numbers for a Measure definition entry! "
            f"Metadata: {measure.metadata}"
        )
        last_end_number = measure.metadata.file_slice.end_line_number
